# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import subprocess

try:
    version = (
        subprocess.check_output(["git", "describe", "--abbrev=0", "--tags"])
        .strip()
        .decode("utf-8")
    )
except Exception:
    print("Could not get version tag. Defaulting to version 0")
    version = "0"

print(version)



# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import re


class _Tests:
    def __init__(self, config, func, files, artifacts) -> None:
        if files is None:
            return

        for key in config:
            attr_name = key

            # Ensure attribute name starts with a letter
            if not attr_name[0].isalpha():
                attr_name = f"test_{attr_name}"

            # Replace non-alphanumeric (and _) characters with _
            attr_name = re.sub(r"[^a-zA-Z0-9_]", "_", attr_name)

            try:
                setattr(
                    self,
                    attr_name,
                    _Test(attr_name, config[key], func, files, artifacts),
                )
            except Exception:
                import traceback

                print(
                    f"Failed to set-up test {key} for function {func.__name__} with traceback: {traceback.format_exc()}"
                )

    def __str__(self) -> str:
        descr = ""
        for key in self.__dict__:
            descr += f"{self.__dict__[key]}\n"
        return descr

    def __repr__(self) -> str:
        return self.__str__()


def _convert_test_files(file_name, files):
    if (
        file_name.startswith("tests/")
        or file_name.startswith("datasets/")
        or file_name.startswith("artifacts/")
    ):
        if file_name.endswith("/"):
            file_name = file_name[:-1]
        return files[file_name]
    return file_name


class _Test:
    def __init__(self, test_name, config, func, files, artifacts) -> None:
        args = config["arguments"].copy()
        for key, vals in args.items():
            if not isinstance(vals, (list, tuple)):
                vals = [vals]

            for i in range(len(vals)):
                val = vals[i]
                if isinstance(val, dict):
                    for k, file_name in val.items():
                        if not isinstance(file_name, str):
                            continue
                        val[k] = _convert_test_files(file_name, files)
                if not isinstance(val, str):
                    continue
                vals[i] = _convert_test_files(val, files)
            args[key] = vals if len(vals) > 1 else vals[0]
        config["arguments"] = args

        self.func = func
        self.name = test_name
        self.config = config

    def run(self, queue=None, dir=None, **kwargs):
        """Run the test

        Parameters
        ----------
        queue : str
            The name of the queue on which to run the test. By default it will be run locally.
            For options check the documentation of the ExecutableDirectory.run method.
        dir : str, Path
            Directory in which to execute the tests. By default they will be executed under
            the /tmp/ folder. However this is usually not accessible over SLURM, therefore
            if queue="slurm" is used, choose a network mounted path in which to execute the tests.
        kwargs : dict
            Additional arguments for the queue execution. For options check the documentation
            of the ExecutableDirectory.run method.
        """
        from playmolecule import PM_APP_ROOT, JobStatus
        from playmolecule._appfiles import _File
        from glob import glob
        import tempfile
        import time
        import inspect
        import os

        with tempfile.TemporaryDirectory(dir=dir) as tmpdir:
            testargs = self.config["arguments"].copy()
            func_sig = inspect.signature(self.func).parameters
            if "outdir" in func_sig:
                testargs["outdir"] = tmpdir
            if "execdir" in func_sig:
                tmpdir = testargs["execdir"]
                if isinstance(tmpdir, _File):
                    tmpdir = tmpdir.path

            print(
                f"Running test '{self.name}' at location {tmpdir} with args {testargs}"
            )

            t = time.time()

            ed = self.func(**testargs)
            ed.run(queue=queue, **kwargs)

            while ed.status not in (JobStatus.COMPLETED, JobStatus.ERROR):
                time.sleep(10)

            for ff in self.config["expected_outputs"]:
                if len(glob(os.path.join(tmpdir, ff))) == 0:
                    raise RuntimeError(
                        f"Test failed. Could not find expected output file: {ff} in {tmpdir}"
                    )
            print(
                f"\n🎉 Test '{self.name}' succeeded in {time.time()-t:.2f} seconds! 🎉"
            )

    def __str__(self) -> str:
        descr = self.config["description"]
        string = f"[{self.name}] '{descr}'\n- Arguments:\n"
        for arg, val in self.config["arguments"].items():
            string += f"  {arg} = {val}\n"
        string += "- Expected outputs:\n"
        for outp in self.config["expected_outputs"]:
            string += f"  {outp}\n"
        return string

    def __repr__(self) -> str:
        return self.__str__()



# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import zipfile
import tempfile
from glob import glob
import logging
import hashlib
import shutil

logger = logging.getLogger(__name__)
SHAREDIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "share")

gcloud = None


def create_gcloud(credentials_file, license_checker=None):
    global gcloud
    gcloud = GCloudStorage(credentials_file, license_checker)


def download_data_location(force=False, default_cred_file=None, license_checker=None):
    import questionary

    if default_cred_file is not None and os.path.exists(default_cred_file):
        create_gcloud(default_cred_file, license_checker)
        return "gcloud"

    image_location_choices = [
        "Google Cloud (Requires internet connection)",
        "Local Directory (For offline installation. Requires large zip file provided by Acellera)",
    ]
    default = image_location_choices[0]

    try:
        image_location = questionary.select(
            "Where do you want to obtain the PlayMolecule images from? :",
            choices=image_location_choices,
            default=default,
            use_shortcuts=True,
        ).unsafe_ask()

        if image_location == image_location_choices[0]:
            # Get google cloud credentials
            gcloud_credentials = questionary.path(
                message="Path to google cloud credentials .json file:",
            ).unsafe_ask()
            shutil.copy(gcloud_credentials, default_cred_file)
            create_gcloud(default_cred_file, license_checker)
            image_location = "gcloud"
        elif image_location == image_location_choices[1]:
            # Get singularity image folder path
            image_location = questionary.path(
                message="Top folder containing the apptainer/singularity images:",
                only_directories=True,
            ).unsafe_ask()
            image_location = "local://" + os.path.abspath(image_location)
    except KeyboardInterrupt:
        raise RuntimeError("Setup interrupted by user. Please start again...")
    return image_location


def update_hashes(target):
    if os.path.isdir(target):
        target = glob(os.path.join(target, "*"))
    else:
        target = [target]

    for tgt in target:
        print(f"Generating MD5 hash for {target}")
        hasher = hashlib.md5()
        with open(tgt, "rb") as infile:
            for chunk in iter(lambda: infile.read(1024 * 1024), b""):
                hasher.update(chunk)
        md5sum = hasher.hexdigest()
        basename = os.path.basename(tgt)
        dirname = os.path.dirname(tgt)
        with open(os.path.join(dirname, f".{basename}.md5sum"), "w") as f:
            f.write(md5sum)


class GCloudStorage:
    def __init__(self, credentials_file, license_checker=None):
        from google.cloud import storage
        from google.oauth2 import service_account

        logger.info("Validating Google Cloud credentials")
        image_bucket_name = "228f801c-064c-46dd-b574-f9f56c1e8d1e"
        cred = service_account.Credentials.from_service_account_file(credentials_file)

        self.storage_client = storage.Client(project=cred._project_id, credentials=cred)
        self.image_bucket = self.storage_client.bucket(image_bucket_name)
        self.license_checker = license_checker
        self.licensed_apps = {}

    def download(self, source, target, _logger=True):
        if _logger:
            print(f"Downloading {source} from google cloud...")
        blob = self.image_bucket.blob(source)
        if not blob.exists():
            # print(f"{source} does not exist.")
            return
        blob.download_to_filename(target)

    def list_files(self, source):
        import subprocess
        from playmolecule import PM_APP_ROOT
        import re

        blobs = self.image_bucket.list_blobs(prefix=source, delimiter="/")
        # Order is important here. The prefixes variable is not populated until we iterate over blobs
        files = [blob.name for blob in blobs if not blob.name.endswith("/")]
        files += list(blobs.prefixes)

        with open(os.path.join(PM_APP_ROOT, "license_path.txt"), "r") as f:
            licence_path = f.read().strip()
        env = os.environ.copy()
        env["ACELLERA_LICENSE_SERVER"] = licence_path

        if self.license_checker:
            # Only show licensed apps to not confuse the user
            licensed_files = []
            for ff in files:
                match = re.search(r"pmws-apps/(\w+)_(v\d+)/", ff)
                if match is None:
                    continue
                app_name = match[1]
                app_version = match[2]
                if app_name in self.licensed_apps:
                    if not self.licensed_apps[app_name]:
                        continue
                else:
                    retcode = subprocess.call(
                        [
                            self.license_checker,
                            "-name",
                            app_name,
                            "-version",
                            app_version[1:],
                        ],
                        stdout=subprocess.DEVNULL,
                        env=env,
                    )
                    self.licensed_apps[app_name] = retcode == 0

                if self.licensed_apps[app_name]:
                    licensed_files.append(ff)
            files = licensed_files
        return sorted(files)


class LocalStorage:
    def __init__(self, data_location, license_checker=None):
        self.data_location = data_location.replace("local://", "")
        self.license_checker = license_checker

    def download(self, source, target, _logger=True):
        if not os.path.exists(os.path.join(self.data_location, source)):
            return
        shutil.copy(os.path.join(self.data_location, source), target)

    def list_files(self, source):
        from glob import glob

        files = []
        for ff in glob(os.path.join(self.data_location, source, "*")):
            if os.path.isdir(ff) and not ff.endswith("/"):
                ff = os.path.relpath(ff, os.path.abspath(self.data_location))
                ff = ff + "/"
            else:
                ff = os.path.relpath(ff, os.path.abspath(self.data_location))
            if self.license_checker:
                raise NotImplementedError(
                    "Add license checking"
                )  # TODO: Check if license exists
            files.append(ff)

        return files


class DataDownloader:
    def __init__(self, data_location, license_checker=None):
        if data_location.startswith("local:///"):
            self.storage = LocalStorage(data_location, license_checker)
        else:
            self.storage = gcloud

    def download(self, source, target, _logger=True):
        if source.endswith("/"):
            directory_mode = True
            files = self.storage.list_files(source)
        else:
            directory_mode = False
            files = [source]

        for ff in files:
            dest = target
            if directory_mode or os.path.isdir(target):
                dest = os.path.join(target, os.path.basename(ff))

            os.makedirs(os.path.dirname(dest), exist_ok=True)
            self.storage.download(ff, dest, _logger=_logger)

    def list_files(self, source):
        return self.storage.list_files(source)


def compare_hashes(source, local_hash_loc, remote_hash_loc):
    from tqdm import tqdm

    download = True
    source_dir = os.path.dirname(remote_hash_loc)

    with tempfile.TemporaryDirectory() as tmpdir:
        # Compare hashes
        tgt = os.path.join(tmpdir, "md5sum")
        source.download(remote_hash_loc, tgt, _logger=False)
        if not os.path.exists(tgt):
            return False

        with open(tgt, "r") as f:
            remote_hash = f.read().strip()

        try:
            if not os.path.exists(local_hash_loc):
                tqdm.write(f"No hashes found locally for {source_dir}. Downloading.")
                return True

            # Download local hash and read
            with open(local_hash_loc, "r") as f:
                local_hash = f.read().strip()

            if local_hash == remote_hash:
                download = False
                tqdm.write(f"Hashes for {source_dir} matched. Skipping download.")
            else:
                tqdm.write(f"Hashes for {source_dir} were different. Downloading.")

        except Exception as e:
            print(f"Failed to get current app hash with error {e}")
            # If there was some crash re-download anyway
            download = True
            pass
    return download


def _update_app_image(datadl, app, appname, outdir):
    hashname = f".{appname}.md5sum"
    manifestname = f"{appname}.json"

    # Compare hashes
    download = compare_hashes(
        datadl, os.path.join(outdir, hashname), f"{app}{hashname}"
    )
    if not download:
        return

    datadl.download(f"{app}{appname}", outdir)
    datadl.download(f"{app}{hashname}", outdir, _logger=False)
    datadl.download(f"{app}{manifestname}", outdir, _logger=False)


def _update_app_examples(datadl, app, outdir):
    examplename = "examples.zip"
    hashname = "examples.md5sum"

    # Compare hashes
    download = compare_hashes(
        datadl, os.path.join(outdir, hashname), f"{app}examples/{hashname}"
    )
    if not download:
        return

    datadl.download(f"{app}examples/{examplename}", outdir)
    datadl.download(f"{app}examples/{hashname}", outdir, _logger=False)

    zipname = os.path.join(outdir, examplename)
    if os.path.exists(zipname):
        with zipfile.ZipFile(zipname, "r") as zip_ref:
            zip_ref.extractall(outdir)
        os.remove(zipname)


def _update_app_tests(datadl, app, outdir):
    testname = "tests.zip"
    hashname = "tests.md5sum"

    outdir = os.path.join(outdir, "files")
    os.makedirs(outdir, exist_ok=True)

    # Compare hashes
    download = compare_hashes(
        datadl, os.path.join(outdir, hashname), f"{app}tests/{hashname}"
    )
    if not download:
        return

    datadl.download(f"{app}tests/{testname}", outdir)
    datadl.download(f"{app}tests/{hashname}", outdir, _logger=False)

    zipname = os.path.join(outdir, testname)
    if os.path.exists(zipname):
        with zipfile.ZipFile(zipname, "r") as zip_ref:
            zip_ref.extractall(outdir)
        os.remove(zipname)


def _update_app_datasets(datadl, app, appname, outdir):
    datasetname = "datasets.zip"
    hashname = "datasets.md5sum"

    outdir = os.path.join(outdir, "files")
    os.makedirs(outdir, exist_ok=True)

    # Compare hashes
    download = compare_hashes(
        datadl, os.path.join(outdir, hashname), f"{app}datasets/{hashname}"
    )
    if not download:
        return

    datadl.download(f"{app}datasets/{datasetname}", outdir)
    datadl.download(f"{app}datasets/{hashname}", outdir, _logger=False)
    print(f"Updating datasets for {appname}")

    zipname = os.path.join(outdir, datasetname)
    if os.path.exists(zipname):
        with zipfile.ZipFile(zipname, "r") as zip_ref:
            zip_ref.extractall(outdir)
        os.remove(zipname)


def _update_app_artifacts(datadl, app, appname, outdir):
    datasetname = "artifacts.zip"
    hashname = "artifacts.md5sum"

    os.makedirs(outdir, exist_ok=True)

    # Compare hashes
    download = compare_hashes(
        datadl, os.path.join(outdir, hashname), f"{app}artifacts/{hashname}"
    )
    if not download:
        return

    datadl.download(f"{app}artifacts/{datasetname}", outdir)
    datadl.download(f"{app}artifacts/{hashname}", outdir, _logger=False)
    print(f"Updating artifacts for {appname}")

    zipname = os.path.join(outdir, datasetname)
    if os.path.exists(zipname):
        with zipfile.ZipFile(zipname, "r") as zip_ref:
            zip_ref.extractall(outdir)
        os.remove(zipname)


def _update_protocols(datadl, app, outdir):
    datasetname = "acellera-protocols.zip"
    hashname = "acellera-protocols.zip.md5sum"

    outdir = os.path.join(outdir, "acellera-protocols")
    os.makedirs(outdir, exist_ok=True)

    # Compare hashes
    download = compare_hashes(
        datadl, os.path.join(outdir, hashname), f"{app}{hashname}"
    )
    if not download:
        return

    datadl.download(f"{app}{datasetname}", outdir)
    datadl.download(f"{app}{hashname}", outdir, _logger=False)
    print("Updating acellera-protocols")

    zipname = os.path.join(outdir, datasetname)
    if os.path.exists(zipname):
        with zipfile.ZipFile(zipname, "r") as zip_ref:
            zip_ref.extractall(outdir)
        os.remove(zipname)


def update_apps(mode="all", automatic=False):
    from playmolecule import PM_APP_ROOT
    from tqdm import tqdm
    import questionary
    import stat

    if PM_APP_ROOT is None:
        raise RuntimeError(f"Could not find PM_APP_ROOT environment variable")

    if PM_APP_ROOT.startswith("docker"):
        from playmolecule._docker import update_docker_apps_from_gcloud

        update_docker_apps_from_gcloud()
        return

    appdir = os.path.join(PM_APP_ROOT, "apps")

    service_acc_json = os.path.join(PM_APP_ROOT, "slpm-service-account.json")
    license_checker = os.path.join(SHAREDIR, "pm-licenses")
    data_location = download_data_location(
        default_cred_file=service_acc_json, license_checker=license_checker
    )

    datadl = DataDownloader(data_location, license_checker)
    apps = datadl.list_files("pmws-apps/")
    apps += ["acellera-protocols/"]

    if len(apps) == 0:
        print(
            "Could not find any apps. You have probably not set up the license server correctly. Set the ACELLERA_LICENSE_SERVER to the port@ip of the license server."
        )
        return

    choices = ["Download all apps", *apps]
    if not automatic:
        try:
            answer = questionary.select(
                "Which app do you want to download?", choices=choices
            ).unsafe_ask()
        except KeyboardInterrupt:
            return
    else:
        answer = "Download all apps"

    if choices.index(answer) == 0:
        answer = apps
    else:
        answer = [answer]

    for app in tqdm(answer, desc="Checking hashes and updating"):
        appname = os.path.basename(os.path.abspath(app))
        name = "_".join(appname.split("_")[:-1])
        version = appname.split("_")[-1]
        outdir = os.path.join(appdir, name.lower(), version)
        os.makedirs(outdir, exist_ok=True)

        artifacts = datadl.list_files(f"{app}artifacts/artifacts.zip")

        with open(os.path.join(outdir, "run.sh"), "w") as f:
            f.write(
                f'#!/bin/bash\nfilename=${{0%.sh}}\njob_dir=`dirname "$(realpath $0)"`\n{PM_APP_ROOT}/apptainer_run.sh {name.lower()}/{version}/{appname} $job_dir $filename'
            )

        if app == "acellera-protocols/":
            _update_protocols(datadl, app, PM_APP_ROOT)
        if mode in ("all", "image"):
            _update_app_image(datadl, app, appname, outdir)
        if len(artifacts) and mode in ("all", "artifacts"):
            _update_app_artifacts(datadl, app, appname, outdir)
        if (
            len(artifacts) == 0
        ):  # TODO: Deprecate this once all apps server artifacts zip
            if mode in ("all", "examples"):
                _update_app_examples(datadl, app, outdir)
            if mode in ("all", "tests"):
                _update_app_tests(datadl, app, outdir)
            if mode in ("all", "datasets"):
                _update_app_datasets(datadl, app, appname, outdir)

    if mode in ("all", "image"):
        _update_app_image(datadl, "pmws-decrypter/", "pm-decrypter", appdir)
        pmdec = os.path.join(appdir, "pm-decrypter")
        st = os.stat(pmdec)
        os.chmod(pmdec, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)



# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
class _Artifacts:
    def __init__(self, artifacts, files_dict) -> None:
        from playmolecule import PM_APP_ROOT
        import os

        searchpaths = [
            ["datasets"],
            ["artifacts"],
            ["files", "datasets"],
            ["files", "artifacts"],
            [],
        ]

        for ds in artifacts:
            path = None
            for sp in searchpaths:
                path = os.path.join(*sp, ds["path"])
                if path in files_dict:
                    break

            if path is None:
                raise RuntimeError(
                    f"Could not find dataset {ds['name']} at path {path}"
                )

            try:
                if "." in ds["name"]:
                    raise RuntimeError(
                        f"Dataset/artifact names cannot include dots in the name. {ds['name']} contains a dot."
                    )
                if not ds["name"][0].isalpha():
                    raise RuntimeError(
                        f"Dataset/artifact names must start with a letter. {ds['name']} does not."
                    )
                setattr(
                    self,
                    ds["name"],
                    _File(ds["name"], files_dict[path].path, ds["description"]),
                )
            except Exception:
                pass

    def __str__(self) -> str:
        descr = ""
        for key in self.__dict__:
            descr += f"{self.__dict__[key]}\n"
        return descr

    def __repr__(self) -> str:
        return self.__str__()


class _File:
    def __init__(self, name, path, description=None) -> None:
        self.name = name
        self.path = path
        self.description = description

    def __str__(self) -> str:
        string = f"[{self.name}] {self.path}"
        if self.description is not None:
            string += f" '{self.description}'"
        return string

    def __repr__(self) -> str:
        return self.__str__()


def _get_app_files(source_dir, manifest=None):
    from playmolecule import PM_APP_ROOT
    from playmolecule._executors._pmbackend import _get_app_files_pmbackend
    from playmolecule._docker import _get_docker_app_files
    from glob import glob
    import os

    if PM_APP_ROOT.startswith("http"):
        return _get_app_files_pmbackend(manifest)

    if PM_APP_ROOT.startswith("docker"):
        return _get_docker_app_files(manifest)

    files = {}

    for ff in glob(os.path.join(source_dir, "**", "*"), recursive=True):
        fname = os.path.relpath(ff, source_dir)
        abspath = os.path.abspath(ff)
        files[fname] = _File(fname, abspath)

    return files



import docker
from google.cloud import artifactregistry_v1
import json
import os


curr_dir = os.path.dirname(os.path.abspath(__file__))


def _get_app_list(project_id="repositories-368911", location="europe-southwest1"):
    # Initialize clients
    ar_client = artifactregistry_v1.ArtifactRegistryClient()

    # 1. Define the parent path for the region
    parent = f"projects/{project_id.strip()}/locations/{location}"

    # 2. List all repositories in that location
    repos_request = artifactregistry_v1.ListRepositoriesRequest(parent=parent)
    repos = ar_client.list_repositories(request=repos_request)

    tags = []

    for repo in repos:
        if repo.format_ == artifactregistry_v1.Repository.Format.DOCKER:
            # 3. List "packages" (the actual images) inside the repository
            # repo.name looks like: projects/PROJ/locations/LOC/repositories/REPO
            pkg_request = artifactregistry_v1.ListPackagesRequest(parent=repo.name)
            packages = ar_client.list_packages(request=pkg_request)

            # The hostname for Artifact Registry
            registry_host = f"{location}-docker.pkg.dev"
            repo_id = repo.name.split("/")[-1]

            # TODO: Remove this hack used during testing
            if repo_id == "acellera-docker-apps":
                continue

            for pkg in packages:
                # Get the image name (last part of the package resource name)
                image_name = pkg.name.split("/")[-1]

                # Full remote path: europe-southwest1-docker.pkg.dev/my-proj/my-repo/my-image
                remote_tag = (
                    f"{registry_host}/{project_id}/{repo_id}/{image_name}:latest"
                )
                tags.append(remote_tag)

    return tags


def update_docker_apps_from_gcloud(
    project_id="repositories-368911", location="europe-southwest1"
):
    from docker import from_env

    docker_client = from_env()

    tags = _get_app_list(project_id, location)
    for tag in tags:
        try:
            print(f"Pulling: {tag} ...")
            docker_client.images.pull(tag)
        except Exception as e:
            print(f"Failed to pull {tag}: {e}")


def _get_app_manifests(prefix):
    """
    Finds local images starting with prefix, runs them to dump a manifest,
    and returns a dictionary of the manifest contents.
    """
    import tempfile
    from pathlib import Path

    client = docker.from_env()

    # 1. Filter local images by prefix
    # We look for tags that start with the prefix (e.g., 'pm-my-app:latest')
    target_images = []
    for img in client.images.list():
        for tag in img.tags:
            if "acellera-docker-apps" in tag:
                continue  # TODO: Remove this hack used during testing
            if tag.startswith(prefix):
                target_images.append(tag)

    manifests = {}

    for image_tag in target_images:
        print(f"--- Processing: {image_tag} ---")

        # 2. Setup temporary host directory for mounting
        with tempfile.TemporaryDirectory() as tmp_dir:
            host_path = Path(tmp_dir)
            container_manifest_dir = "/manifest"
            filename = "manifest.json"

            try:
                # 3. Run the container
                # command: The argument to pass to the entrypoint
                # volumes: Bind mount the temp dir to /manifest
                # remove: Clean up the container immediately after exit
                client.containers.run(
                    image=image_tag,
                    command=f"--dump-manifest {container_manifest_dir}/{filename}",
                    volumes={
                        str(host_path.resolve()): {
                            "bind": container_manifest_dir,
                            "mode": "rw",
                        },
                        "/etc/passwd": {"bind": "/etc/passwd", "mode": "ro"},
                        "/etc/group": {"bind": "/etc/group", "mode": "ro"},
                    },
                    remove=True,
                    user=f"{os.getuid()}:{os.getgid()}",
                    extra_hosts={"host.docker.internal": "host-gateway"},
                    environment={
                        "ACELLERA_LICENCE_SERVER": "27000@host.docker.internal",
                    },
                )

                # 4. Read the file from the host's temp directory
                manifest_file = host_path / filename
                if manifest_file.exists():
                    with open(manifest_file, "r") as f:
                        manifests[image_tag] = json.load(f)
                else:
                    raise RuntimeError(f"No manifest file found for {image_tag}")

            except Exception as e:
                raise RuntimeError(f"Failed to run {image_tag}: {e}")

    return manifests


def _get_apps(project_id="repositories-368911", location="europe-southwest1"):
    cache_folder = os.path.join(os.path.expanduser("~"), ".cache", "playmolecule")
    os.makedirs(cache_folder, exist_ok=True)

    prefix = f"{location}-docker.pkg.dev/{project_id}"
    manifests = _get_app_manifests(prefix)
    apps = {}

    for image_tag, manifest in manifests.items():
        container_config = manifest["container_config"]
        app_name = container_config["name"].lower()
        version = container_config["version"]

        with open(os.path.join(curr_dir, "share", "docker_run.sh"), "r") as f:
            run_sh = f.read()

        run_sh = run_sh.replace("{docker_container_name}", f'"{image_tag}"')

        cache_run_sh = os.path.join(cache_folder, f"run_{app_name}_{version}.sh")
        with open(cache_run_sh, "w") as f:
            f.write(run_sh)

        if app_name not in apps:
            apps[app_name] = {}
        apps[app_name][f"v{version}"] = {
            "manifest": manifest,
            "appdir": None,
            "run.sh": cache_run_sh,
            "container_image": image_tag,
        }

    return apps


def _set_docker_root_gcloud():
    from playmolecule.apps import _manifest_to_func, _link_latest_version
    from natsort import natsorted

    app_manifests = _get_apps()

    for appname in app_manifests:
        func_names = _manifest_to_func(appname, app_manifests[appname])
        _link_latest_version(
            appname, natsorted(app_manifests[appname].keys())[-1], func_names
        )


def _get_docker_app_files(manifest):
    from playmolecule._appfiles import _File

    files = {}
    for name, fullpath in manifest["files"].items():
        files[name] = _File(name, fullpath)

    return files


if __name__ == "__main__":
    PROJECT = "repositories-368911"
    REGION = "europe-southwest1"

    results = update_docker_apps_from_gcloud(PROJECT, REGION)
    print("\nSuccessfully pulled and tagged:")
    for c in results:
        print(f" - {c}")

    results = _get_apps(prefix="pm-")
    print("\n" + "=" * 30)
    print("EXTRACTED MANIFESTS")
    print("=" * 30)
    print(json.dumps(results, indent=2))

    # docker run --add-host=host.docker.internal:host-gateway -e ACELLERA_LICENCE_SERVER=27000@host.docker.internal pm-proteinprepare:latest



import os

old_header = r"""# (c) 2015-2012 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
"""

new_header = r"""# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
"""

exclusions = "playmolecule/_version.py"

for root, dirs, files in os.walk("playmolecule"):
    for i, fname in enumerate(
        [
            os.path.join(root, file)
            for file in files
            if not any(exclusion in root for exclusion in exclusions)
        ]
    ):
        if fname in exclusions:
            continue

        if fname.endswith((".py", ".sh", ".jinja")):
            try:
                print(f"-------------------------- {fname} ---------------------------")
                with open(fname) as f:
                    print("".join(f.readlines()[:6]))
                with open(fname) as f:
                    text = f.read()

                if old_header in text:
                    text = text.replace(old_header, new_header)
                else:
                    text = new_header + text
                print("XXXXXXXXXXXXXXXX")
                print("\n".join(text.split("\n")[:10]))
                with open(fname, "w") as fout:
                    fout.write(text)
            except Exception as e:
                print(f"ERROR: {e}")



def __find_protocols(root_dir, parent_module="playmolecule"):
    import os

    prot_dir = os.path.join(root_dir, "acellera-protocols")

    if os.path.exists(root_dir) and os.path.exists(prot_dir):
        from glob import glob
        import importlib
        import sys

        sys.path.insert(0, prot_dir)

        for file in glob(os.path.join(prot_dir, "**", "*.py"), recursive=True):
            if file.endswith("__init__.py"):
                continue
            rel_path = os.path.relpath(file[:-3], prot_dir)
            mod_name = rel_path.replace(os.path.sep, ".")
            parts = mod_name.split(".")

            for i in range(len(parts)):
                submod = ".".join(parts[: i + 1])
                sys.modules[parent_module + "." + submod] = importlib.import_module(
                    submod, package=parent_module
                )

            # Append tutorials to the docs of the protocol
            dirname = os.path.dirname(file)
            pieces = rel_path.split(os.path.sep)
            # Check if the loaded module is the actual protocol file
            if len(pieces) == 4 and pieces[1] == pieces[3]:
                # Check if there are files/tutorials
                nb_tuts = glob(os.path.join(dirname, "files", "tutorials", "*.ipynb"))
                if len(nb_tuts):
                    # Modify the docs of the protocol
                    main_func = getattr(
                        sys.modules[parent_module + "." + submod], pieces[1]
                    )
                    main_func.__doc__ = (
                        main_func.__doc__
                        + "\n\nNotes\n-----\nTutorials are available for this protocol:\n\n"
                        + "\n".join([f"    - {t}" for t in nb_tuts])
                    )

        return sys.modules[parent_module + ".protocols"]
    return None



# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
import sys
import logging.config
from playmolecule.apps import _set_root, JobStatus, ExecutableDirectory
from playmolecule._executors._slurm import slurm_mps
from playmolecule._update import update_apps
from importlib.metadata import version as __version
from importlib.metadata import PackageNotFoundError
from playmolecule._protocols import __find_protocols

# from importlib.resources import files

try:
    __version__ = __version("playmolecule")
except PackageNotFoundError:
    pass

dirname = os.path.dirname(__file__)

try:
    logging.config.fileConfig(
        os.path.join(dirname, "logging.ini"), disable_existing_loggers=False
    )
except Exception:
    print("playmolecule: Logging setup failed")


logger = logging.getLogger(__name__)
PM_APP_ROOT = os.environ.get("PM_APP_ROOT", None)
PM_SKIP_SETUP = os.environ.get("PM_SKIP_SETUP", False)
if not PM_SKIP_SETUP and PM_APP_ROOT is None:
    raise RuntimeError(
        "Could not find environment variable PM_APP_ROOT. Please set the variable to set the path to the app root."
    )

if PM_APP_ROOT is not None and not PM_SKIP_SETUP:
    # Necessary for PMWS URL but also good for paths
    while PM_APP_ROOT[-1] == "/":
        PM_APP_ROOT = PM_APP_ROOT[:-1]
    _set_root(PM_APP_ROOT)


# Requires PM_APP_ROOT to be set, otherwise we get circular import errors
from playmolecule._executors._pmbackend import login, logout


def describe_apps(as_dict=False):
    from playmolecule.apps import _function_dict

    app_dict = {}
    sorted_keys = sorted(_function_dict.keys())
    for func_path in sorted_keys:
        func = _function_dict[func_path]
        if "name" in func.__manifest__:
            name = func.__manifest__["name"]
        else:
            name = func_path.split(".")[-1]
        app_dict[func_path] = {"description": func.__doc__.strip().split("\n")[0]}
        if not as_dict:
            print(name, func_path)
            desc = func.__doc__.strip().split("\n")[0]
            print(f"    {desc}")

    if as_dict:
        return app_dict


module = sys.modules[__name__]
setattr(module, "protocols", None)

# Add the acellera-protocols folder as a submodule
if not PM_SKIP_SETUP:
    setattr(module, "protocols", __find_protocols(PM_APP_ROOT, parent_module=__name__))



def submit_slurm(dirname, runsh, execution_resources, **kwargs):
    from jobqueues.slurmqueue import SlurmQueue

    sl = SlurmQueue()
    sl.runscript = runsh

    if execution_resources is not None:
        # Set app defaults
        for arg in execution_resources:
            setattr(sl, arg, execution_resources[arg])

    # Set user-specified arguments
    for arg in kwargs:
        setattr(sl, arg, kwargs[arg])

    sl.submit(dirname)
    return sl


def get_slurm_status(slurmq):
    from jobqueues.simqueue import QueueJobStatus
    from playmolecule.apps import JobStatus

    mapping = {
        QueueJobStatus.RUNNING: JobStatus.RUNNING,
        QueueJobStatus.FAILED: JobStatus.ERROR,
        QueueJobStatus.CANCELLED: JobStatus.ERROR,
        QueueJobStatus.OUT_OF_MEMORY: JobStatus.ERROR,
        QueueJobStatus.TIMEOUT: JobStatus.ERROR,
        QueueJobStatus.COMPLETED: JobStatus.COMPLETED,
        QueueJobStatus.PENDING: JobStatus.WAITING_INFO,
        None: JobStatus.WAITING_INFO,
    }
    info = slurmq.jobInfo()
    if info is None:
        return JobStatus.WAITING_INFO
    return mapping[info[list(info.keys())[0]]["state"]]


def slurm_mps(exec_dirs, **kwargs):
    """Submit a list of ExecutableDirectories to SLURM as a single MPS job.

    This means that all jobs submitted will be executed on the same GPU

    Parameters
    ----------
    exec_dirs : list[ExecutableDirectory]
        An iterable of ExecutableDirectory objects
    partition : str or list of str
        The queue (partition) or list of queues to run on. If list, the one offering earliest initiation will be used.
    jobname : str
        Job name (identifier)
    priority : str
        Job priority
    ncpu : int
        Number of CPUs to use for a single job
    ngpu : int
        Number of GPUs to use for a single job
    memory : int
        Amount of memory per job (MiB)
    gpumemory : int
        Only run on GPUs with at least this much memory. Needs special setup of SLURM. Check how to define gpu_mem on
        SLURM.
    walltime : int
        Job timeout (s)
    mailtype : str
        When to send emails. Separate options with commas like 'END,FAIL'.
    mailuser : str
        User email address.
    outputstream : str
        Output stream.
    errorstream : str
        Error stream.
    nodelist : list
        A list of nodes on which to run every job at the *same time*! Careful! The jobs will be duplicated!
    exclude : list
        A list of nodes on which *not* to run the jobs. Use this to select nodes on which to allow the jobs to run on.
    envvars : str
        Envvars to propagate from submission node to the running node (comma-separated)
    prerun : list
        Shell commands to execute on the running node before the job (e.g. loading modules)

    Examples
    --------
    >>> ed1 = kdeep(outdir="test1", pdb=apps.kdeep.files["tests/10gs_protein.pdb"], sdf=apps.kdeep.files["tests/10gs_ligand.sdf"], modelfile=kdeep.datasets.default)
    >>> ed2 = kdeep(outdir="test2", dataset=apps.kdeep.files["tests/dataset.zip"], modelfile=kdeep.datasets.default)
    >>> slurm_mps([ed1, ed2], partition="normalGPU", ncpu=1, ngpu=1)
    """
    from jobqueues.slurmqueue import SlurmQueue

    sl = SlurmQueue()

    if exec_dirs[0].execution_resources is not None:
        # Set app defaults
        for arg in exec_dirs[0].execution_resources:
            setattr(sl, arg, exec_dirs[0].execution_resources[arg])

    # Set user-specified arguments
    for arg in kwargs:
        setattr(sl, arg, kwargs[arg])

    sl.submit(
        dirs=[ed.dirname for ed in exec_dirs],
        runscripts=[ed.runsh for ed in exec_dirs],
        nvidia_mps=True,
    )
    return sl



# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from playmolecule.apps import JobStatus
from playmolecule import PM_APP_ROOT
import requests
import logging
import json
import os

logger = logging.getLogger(__name__)

# --- Global State ---
# This persists as long as the python script is running.
# We initialize it immediately so it's ready to use.
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

_internal_session = requests.Session()

# Configure connection pooling + retries.
# This makes status polling resilient to transient disconnects (e.g. backend/proxy
# closing an idle keep-alive connection).
_retry_strategy = Retry(
    total=5,
    connect=5,
    read=5,
    status=5,
    backoff_factor=0.5,
    status_forcelist=(429, 500, 502, 503, 504),
    allowed_methods=frozenset({"GET", "POST"}),
    respect_retry_after_header=True,
)
_adapter = HTTPAdapter(
    max_retries=_retry_strategy,
    pool_connections=10,
    pool_maxsize=10,
    pool_block=False,
)
_internal_session.mount("http://", _adapter)
_internal_session.mount("https://", _adapter)

# Default timeout: (connect, read). Avoids hanging forever on backend issues.
_DEFAULT_TIMEOUT = (5, 30)

_base_url = PM_APP_ROOT
# Try to read headers from environment variable PM_BACKEND_HEADERS
PM_BACKEND_HEADERS = os.environ.get("PM_BACKEND_HEADERS", None)
if PM_BACKEND_HEADERS is not None:
    _headers = json.loads(PM_BACKEND_HEADERS)
else:
    _headers = {}


def login(username, password):
    """
    Authenticates and stores cookies in the global module session.
    """
    if not _base_url.startswith("http"):
        raise RuntimeError(f"Invalid backend URL: {_base_url}")

    # Get the unauthenticated CSRF token
    response = _internal_session.get(
        f"{_base_url}/auth/csrf", headers=_headers, timeout=_DEFAULT_TIMEOUT
    )
    response.raise_for_status()
    csrf_token = response.json()["csrf_token"]

    headers = _headers.copy()
    headers["X-CSRF-Token"] = csrf_token

    # This updates _internal_session.cookies automatically
    response = _internal_session.post(
        f"{_base_url}/auth/login",
        data={"username": username, "password": password},
        headers=headers,
        timeout=_DEFAULT_TIMEOUT,
    )
    response.raise_for_status()

    print(f"Logged in as {username}")


def logout():
    """
    Clears the cookies from the global session.
    """
    _internal_session.cookies.clear()
    print("Logged out.")


def _get_apps():
    print("Querying", f"{_base_url}/apps/manifests")
    rsp = _internal_session.get(
        f"{_base_url}/apps/manifests", headers=_headers, timeout=_DEFAULT_TIMEOUT
    )
    if rsp.status_code != 200:
        raise RuntimeError(f"Failed to get apps from {_base_url}: {rsp.text}")

    apps = {}
    for app_module, manifest in rsp.json().items():
        app_name = app_module.split(".")[0]
        if app_name not in apps:
            apps[app_name] = {}

        version = manifest.get("version")
        if version is None:
            try:
                version = manifest["container_config"]["version"]
            except:
                version = "1"

        apps[app_name][f"v{int(float(version))}"] = {
            "manifest": manifest,
            "appdir": None,
            "run.sh": None,
        }

    return apps


def _set_root_pmbackend():
    from playmolecule.apps import _manifest_to_func, _link_latest_version
    from natsort import natsorted

    app_manifests = _get_apps()

    for appname in app_manifests:
        func_names = _manifest_to_func(appname, app_manifests[appname])
        _link_latest_version(
            appname, natsorted(app_manifests[appname].keys())[-1], func_names
        )


def _pmbackend_status(dirname, job_id):
    import safezipfile
    import io

    # Query job metadata

    try:
        headers = _headers.copy()
        headers["Connection"] = "close"
        res = _internal_session.get(
            f"{_base_url}/jobs?prefix={job_id}",
            timeout=_DEFAULT_TIMEOUT,
            headers=headers,
        )
    except requests.exceptions.RequestException as e:
        # Typical root cause: proxy/backend closes an idle keep-alive connection
        # and the client tries to reuse it (RemoteDisconnected).
        logger.warning(f"PM backend status poll failed for job '{job_id}': {e}")
        return JobStatus.WAITING_INFO
    res.raise_for_status()
    jobs = res.json()
    if not isinstance(jobs, list) or len(jobs) != 1:
        return JobStatus.WAITING_INFO

    metadata = jobs[0]
    status = {0: "WAITING_INFO", 1: "RUNNING", 2: "COMPLETED", 3: "ERROR"}[
        int(metadata["status"])
    ]
    # On error, return combined error text (no files)
    if status == "ERROR":
        res = _internal_session.get(
            f"{_base_url}/files?prefix={job_id}/",
            timeout=_DEFAULT_TIMEOUT,
            headers=_headers,
        )
        files = res.json()
        errf = None
        outf = None
        for f in files.values():
            if f["name"].startswith("slurm."):
                if f["name"].endswith(".err"):
                    errf = f["uri"]
                elif f["name"].endswith(".out"):
                    outf = f["uri"]

        errres = _internal_session.get(
            f"{_base_url}/file/{job_id}{errf}",
            timeout=_DEFAULT_TIMEOUT,
            headers=_headers,
        )
        outres = _internal_session.get(
            f"{_base_url}/file/{job_id}{outf}",
            timeout=_DEFAULT_TIMEOUT,
            headers=_headers,
        )
        combined_error = f"{outres.content}{errres.content}"
        return JobStatus.ERROR

    # While running/waiting: no files
    if status in ("WAITING_INFO", "RUNNING"):
        return JobStatus.RUNNING

    # Completed. Download files
    res = _internal_session.get(
        f"{_base_url}/file/{job_id}/",
        timeout=(5, 360),  # Longer read timeout for file downloads
        headers=_headers,
    )
    with safezipfile.ZipFile(io.BytesIO(res.content)) as zf:
        zf.extractall(dirname, max_files=1e9, max_file_size=1e11, max_total_size=1e12)

    return JobStatus.COMPLETED


def _get_inputs(arguments, manifest, app_files, slpm_path):
    from playmolecule.apps import _validators
    from playmolecule._appfiles import _File

    files = []
    file_handles = []  # Keep track of file handles to close later

    if "outdir" in arguments:
        del arguments["outdir"]
    if "scratchdir" in arguments:
        del arguments["scratchdir"]
    if "execdir" in arguments:
        del arguments["execdir"]

    # Validate arg types and copy Path arguments to folder
    original_paths = {}
    for arg in manifest["params"]:
        name = arg["name"]
        argtype = arg["type"]
        nargs = None if "nargs" not in arg else arg["nargs"]

        if name in ("outdir", "scratchdir", "execdir"):
            continue

        vals = arguments[name]
        if nargs is None and isinstance(vals, (list, tuple)):
            raise RuntimeError(
                f"Argument '{name}' was passed value '{vals}' which is of type '{type(vals).__name__}'. Was expecting a single value of type '{argtype}'."
            )
        if not isinstance(vals, (list, tuple)):
            vals = [vals]

        # Validate type
        if argtype in _validators:
            validator = _validators[argtype]
            for val in vals:
                if val is not None and not isinstance(val, validator):
                    raise RuntimeError(
                        f"Argument '{name}' was passed value '{val}' which is of type '{type(val).__name__}'. Was expecting value of type '{argtype}'."
                    )
        else:
            logger.warning(
                f"Could not find validator for type: {arg['type']}. Please notify the PM developers."
            )

        # Open files for Path-type arguments
        if argtype == "Path" and name in arguments:
            newvals = []
            for val in vals:
                if val is None or (isinstance(val, str) and val == ""):
                    continue

                if isinstance(val, str) and val.startswith("app://files"):
                    newvals.append(val)
                    continue

                if isinstance(val, _File):
                    newvals.append(val.path)
                    continue  # Don't copy artifacts

                val = os.path.abspath(val)
                filename = os.path.basename(val)
                if os.path.isdir(val):
                    raise RuntimeError(
                        f"Directory arguments are not supported for PM backend: {val}"
                    )
                fh = open(val, "rb")
                file_handles.append(fh)  # Track for cleanup
                files.append((name, (filename, fh, "application/octet-stream")))
                newvals.append(filename)

            if len(newvals) == 0:
                arguments[name] = None
            elif len(newvals) == 1:
                arguments[name] = newvals[0]
            else:
                arguments[name] = newvals

    return {
        "arguments": arguments,
        "slpm_path": slpm_path,
        "files": files,
        "file_handles": file_handles,
    }


def _run_on_pmbackend(input_dict, job_id="", prefix="default-project/", _logger=True):
    if "access_token" not in _internal_session.cookies:
        raise RuntimeError("Access token not found. Please login to the PM backend.")

    file_handles = input_dict.get("file_handles", [])

    try:
        headers = _headers.copy()
        headers["X-CSRF-Token"] = _internal_session.cookies.get("csrf_token")
        res = _internal_session.post(
            f"{_base_url}/apps/{input_dict['slpm_path']}/run?job_dir={prefix}{job_id}",
            data={"input": json.dumps(input_dict["arguments"])},
            files=input_dict["files"],
            headers=headers,
            timeout=(10, 120),
        )

        if res.status_code != 200:
            raise RuntimeError(f"Failed to run job on {_base_url}: {res.text}")
        return res.json()["job_id"]
    finally:
        # Always close file handles, even if an error occurs
        for fh in file_handles:
            try:
                fh.close()
            except Exception as e:
                logger.warning(f"Failed to close file handle: {e}")


def _get_app_files_pmbackend(manifest):
    from playmolecule._appfiles import _File

    files = manifest.get("files", {})
    for name, description in files.items():
        while name.endswith("/"):
            name = name[:-1]
        path = f"app://files/{name}"
        files[name] = _File(name, path, description)

    return files



# (c) 2015-2023 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from glob import glob
import os
from jinja2 import Environment, PackageLoader, select_autoescape
from playmolecule._appfiles import _File, _Artifacts, _get_app_files
from playmolecule._tests import _Tests
import shutil

# Do not remove unused imports. They are used in the jinja file probably
from pathlib import Path
import stat
import logging
import enum


logger = logging.getLogger(__name__)


class KWARGS(dict):
    pass


_function_dict = {}


env = Environment(
    loader=PackageLoader("playmolecule", "share"),
    autoescape=select_autoescape(["*"]),
)


@enum.unique
class JobStatus(enum.IntEnum):
    """Job status codes describing the current status of a job

    * WAITING_INFO : Waiting for status from the job. Job has not started yet computation.
    * RUNNING : Job is currently running
    * COMPLETED : Job has successfully completed
    * ERROR : Job has exited with an error
    """

    WAITING_INFO = 0
    RUNNING = 1
    COMPLETED = 2
    ERROR = 3

    def describe(self):
        codes = {0: "Waiting info", 1: "Running", 2: "Completed", 3: "Error"}
        return codes[self.value]

    def __str__(self):
        return self.describe()


JOB_TIMEOUT = 60  # Timeout after which if there is no newer date in .pm.alive we consider the job dead


class ExecutableDirectory:
    """Executable directory class.

    All app functions will create a folder and return to you an `ExecutableDirectory` object.
    This is a self-contained directory including all app input files which can be executed either locally or on a cluster.
    If it's not executed locally make sure the directory can be accessed from all machines
    in the cluster (i.e. is located on a shared filesystem).
    """

    def __init__(
        self, dirname, inputs_dir=None, _execution_resources=None, _input_json=None
    ) -> None:
        self.dirname = dirname

        if inputs_dir is None:
            # Find all the run_ folders in the directory and use the most recent one
            run_folders = [
                f
                for f in os.listdir(dirname)
                if f.startswith("run_") and os.path.isdir(os.path.join(dirname, f))
            ]
            run_folders.sort(key=lambda x: os.path.getmtime(os.path.join(dirname, x)))
            inputs_dir = os.path.join(dirname, run_folders[-1])

        self.runsh = os.path.basename(inputs_dir) + ".sh"
        self.inputs_dir = inputs_dir

        self.execution_resources = _execution_resources
        self.slurmq = None
        self._input_json = _input_json

    @property
    def status(self):
        """Returns current status of the ExecutableDirectory

        Examples
        --------
        >>> ed = proteinprepare(outdir="test", pdbid="3ptb")
        >>> ed.slurm(ncpu=1, ngpu=0)
        >>> print(ed.status)
        """
        from playmolecule import PM_APP_ROOT
        from playmolecule._executors._pmbackend import _pmbackend_status
        import datetime
        import json

        if PM_APP_ROOT.startswith("http"):
            if self._job_id is None:
                raise RuntimeError("Job ID is not set. Please run the job first.")
            return _pmbackend_status(self.dirname, self._job_id)

        # If the container reported completion or error
        # TODO: Remove the checks in dirname once all apps have been updated to write in the sentinel_dir
        if os.path.exists(os.path.join(self.inputs_dir, ".pm.done")) or os.path.exists(
            os.path.join(self.dirname, ".pm.done")
        ):
            return JobStatus.COMPLETED
        elif os.path.exists(os.path.join(self.inputs_dir, ".pm.err")) or os.path.exists(
            os.path.join(self.dirname, ".pm.err")
        ):
            return JobStatus.ERROR

        # If the container is reporting aliveness
        heartbeat = os.path.join(self.inputs_dir, ".pm.alive")
        if not os.path.exists(heartbeat):
            heartbeat = os.path.join(self.dirname, ".pm.alive")
        if os.path.exists(heartbeat):
            with open(heartbeat, "r") as f:
                timestamp_str = f.read().strip()
                timestamp = None

                if len(timestamp_str):
                    try:
                        timestamp = datetime.datetime.fromisoformat(timestamp_str)
                    except Exception:
                        logger.error(f"Malformed timestamp in {heartbeat}")

                if timestamp is not None:
                    diff = datetime.datetime.now() - timestamp
                    if diff > datetime.timedelta(seconds=JOB_TIMEOUT):
                        return JobStatus.ERROR
                    else:
                        return JobStatus.RUNNING

        # If it was submitted to SLURM handle the state here
        if self.slurmq is not None:
            from playmolecule._executors._slurm import get_slurm_status

            return get_slurm_status(self.slurmq)

        # Alternatively check if expected outputs exist
        outputs = os.path.join(self.inputs_dir, "expected_outputs.json")
        if os.path.exists(outputs):
            with open(outputs, "r") as f:
                outputs = json.load(f)

            for outf in outputs:
                if len(glob(os.path.join(self.dirname, outf))) == 0:
                    return JobStatus.RUNNING
            else:
                return JobStatus.COMPLETED

        logger.warning(
            f"Could not yet determine job status for directory {self.dirname}. The job might have not started running yet."
        )
        return JobStatus.WAITING_INFO

    def run(self, queue=None, verbose=True, **kwargs):
        """Execute the directory locally

        If no queue is specified it will run the job locally.

        Examples
        --------
        >>> ed = proteinprepare(outdir="test", pdbid="3ptb")
        >>> ed.run()

        Specifying a queue

        >>> ed.run(queue="slurm", partition="normalCPU", ncpu=3, ngpu=1)

        Alternative syntax for

        >>> ed.slurm(partition="normalCPU", ncpu=3, ngpu=1)
        """
        from playmolecule import PM_APP_ROOT
        from playmolecule._executors._pmbackend import _run_on_pmbackend
        import json

        if queue is None:
            # Check if the queue is specified in the environment variable PM_QUEUE_CONFIG
            PM_QUEUE_CONFIG = os.environ.get("PM_QUEUE_CONFIG", None)
            if PM_QUEUE_CONFIG is not None:
                qconf = json.loads(PM_QUEUE_CONFIG)
                queue = qconf["queue"].lower()
                del qconf["queue"]
                if queue == "slurm":
                    # If it's SLURM try to pick the appropriate partition depending on the number of GPUs
                    needs_gpu = self.execution_resources.get("ngpu", 0) > 0
                    partition = (
                        qconf["gpu_partition"] if needs_gpu else qconf["cpu_partition"]
                    )
                    del qconf["gpu_partition"]
                    del qconf["cpu_partition"]
                    qconf["partition"] = partition
                kwargs = qconf

        self.slurmq = None  # New execution. Set to None

        if PM_APP_ROOT.startswith("http"):
            self._job_id = _run_on_pmbackend(self._input_json, **kwargs)
            return

        if queue is None:
            import subprocess

            logfile = os.path.join(self.dirname, self.runsh.replace(".sh", ".log"))

            with open(logfile, "w") as logfile:
                process = subprocess.Popen(
                    ["bash", self.runsh],
                    cwd=self.dirname,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                )
                for line in process.stdout:
                    if verbose:
                        print(line, end="")  # Write to terminal
                    logfile.write(line)  # Write to file
                process.wait()
        else:
            if queue.lower() == "slurm":
                self.slurm(**kwargs)
            elif queue.lower() == "pmws":
                _run_on_pmbackend(self.inputs_dir, **kwargs)

    def slurm(self, **kwargs):
        """Submit simulations to SLURM cluster

        Parameters
        ----------
        partition : str or list of str
            The queue (partition) or list of queues to run on. If list, the one offering earliest initiation will be used.
        jobname : str
            Job name (identifier)
        priority : str
            Job priority
        ncpu : int
            Number of CPUs to use for a single job
        ngpu : int
            Number of GPUs to use for a single job
        memory : int
            Amount of memory per job (MiB)
        gpumemory : int
            Only run on GPUs with at least this much memory. Needs special setup of SLURM. Check how to define gpu_mem on
            SLURM.
        walltime : int
            Job timeout (s)
        mailtype : str
            When to send emails. Separate options with commas like 'END,FAIL'.
        mailuser : str
            User email address.
        outputstream : str
            Output stream.
        errorstream : str
            Error stream.
        nodelist : list
            A list of nodes on which to run every job at the *same time*! Careful! The jobs will be duplicated!
        exclude : list
            A list of nodes on which *not* to run the jobs. Use this to select nodes on which to allow the jobs to run on.
        envvars : str
            Envvars to propagate from submission node to the running node (comma-separated)
        prerun : list
            Shell commands to execute on the running node before the job (e.g. loading modules)

        Examples
        --------
        >>> ed = proteinprepare(outdir="test", pdbid="3ptb")
        >>> ed.slurm(partition="normalCPU", ncpu=1, ngpu=0)
        """
        from playmolecule._executors._slurm import submit_slurm

        self.slurmq = submit_slurm(
            self.dirname, self.runsh, self.execution_resources, **kwargs
        )


_validators = {
    "str": str,
    "Path": (str, Path, _File),
    "bool": bool,
    "int": (int, float),
    "float": (int, float),
    "dict": dict,
}


def _inner_function(
    args: dict,
    function_name: str,
    function_resources: dict,
    func_manifest: dict,
    slpm_path: str,
    files: dict,
    run_sh: str,
):
    from playmolecule.apps import _write_inputs, ExecutableDirectory
    from playmolecule import PM_APP_ROOT
    from datetime import datetime
    import uuid
    import shutil
    import os

    assert (
        "outdir" in args or "execdir" in args
    ), "Functions must accept either 'outdir' or 'execdir'"

    if "outdir" in args:
        outdir = args["outdir"]
        args["outdir"] = "."
    elif "execdir" in args:
        if isinstance(args["execdir"], _File):
            args["execdir"] = args["execdir"].path
        outdir = args["execdir"]
        args["execdir"] = "."

    if PM_APP_ROOT.startswith("http"):
        from playmolecule._executors._pmbackend import _get_inputs

        input_dict = _get_inputs(args, func_manifest, files, slpm_path)
        ed = ExecutableDirectory(
            dirname=outdir,
            inputs_dir=outdir,
            _execution_resources=function_resources,
            _input_json=input_dict,
        )
        BLOCKING = os.getenv("PM_BLOCKING", "0") == "1"
        if not BLOCKING:
            return ed
        else:
            import time

            ed.run()
            time.sleep(2)
            while ed.status not in (JobStatus.COMPLETED, JobStatus.ERROR):
                time.sleep(3)
            return None

    if "scratchdir" in args:
        if args["scratchdir"] is None:
            args["scratchdir"] = os.path.join(outdir, "scratch")
        os.makedirs(args["scratchdir"], exist_ok=True)

    now = datetime.now()
    identifier = f"_{now.strftime(r'%d_%m_%Y_%H_%M')}_{uuid.uuid4().hex[:8]}"
    inputs_dir = os.path.join(outdir, f"run{identifier}")
    _write_inputs(
        outdir, inputs_dir, args, func_manifest, files, function_name, slpm_path
    )

    target_run_sh = None
    if run_sh is not None and len(run_sh):
        target_run_sh = os.path.join(outdir, f"run{identifier}.sh")
        with open(target_run_sh, "w") as f:
            f.write(run_sh)
        st = os.stat(target_run_sh)
        os.chmod(target_run_sh, st.st_mode | stat.S_IEXEC)

    return ExecutableDirectory(
        dirname=outdir, inputs_dir=inputs_dir, _execution_resources=function_resources
    )


def _write_inputs(
    write_dir, inputs_dir, arguments, manifest, app_files, function, slpm_path
):
    import json

    os.makedirs(inputs_dir, exist_ok=True)

    if "outputs" in manifest:
        with open(os.path.join(inputs_dir, "expected_outputs.json"), "w") as f:
            json.dump(manifest["outputs"], f)

    # Validate arg types and copy Path arguments to folder
    original_paths = {}
    for arg in manifest["params"]:
        name = arg["name"]
        argtype = arg["type"]
        nargs = None if "nargs" not in arg else arg["nargs"]

        vals = arguments[name]
        if nargs is None and isinstance(vals, (list, tuple)):
            raise RuntimeError(
                f"Argument '{name}' was passed value '{vals}' which is of type '{type(vals).__name__}'. Was expecting a single value of type '{argtype}'."
            )
        if not isinstance(vals, (list, tuple)):
            vals = [vals]

        # Validate type
        if argtype in _validators:
            validator = _validators[argtype]
            for val in vals:
                if val is not None and not isinstance(val, validator):
                    raise RuntimeError(
                        f"Argument '{name}' was passed value '{val}' which is of type '{type(val).__name__}'. Was expecting value of type '{argtype}'."
                    )
        else:
            logger.warning(
                f"Could not find validator for type: {arg['type']}. Please notify the PM developers."
            )

        # Copy Path-type arguments to folder
        if (
            argtype == "Path"
            and name in arguments
            and name not in ("outdir", "scratchdir", "execdir")
        ):
            newvals = []
            for val in vals:
                if val is None or (isinstance(val, str) and val == ""):
                    continue

                if isinstance(val, str) and val.startswith("app://files"):
                    val = app_files[val.replace("app://files/", "")]

                if isinstance(val, _File):
                    newvals.append(val.path)
                    continue  # Don't copy artifacts

                val = os.path.abspath(val)

                # Deduplicate file names
                outname = os.path.join(inputs_dir, os.path.basename(val))
                if os.path.exists(outname) and val != original_paths[outname]:
                    i = 0
                    while os.path.exists(outname):
                        parts = os.path.splitext(os.path.basename(val))
                        outname = os.path.join(inputs_dir, f"{parts[0]}_{i}{parts[1]}")
                        i += 1

                original_paths[outname] = val

                if "PM_SYMLINK" in os.environ:
                    os.symlink(val, outname)
                else:
                    if os.path.isdir(val):
                        shutil.copytree(val, outname)
                    else:
                        shutil.copy(val, outname)
                newvals.append(os.path.relpath(outname, write_dir))

            if len(newvals) == 0:
                arguments[name] = None
            elif len(newvals) == 1:
                arguments[name] = newvals[0]
            else:
                arguments[name] = newvals

        # Special case for the HTMD app which takes arbitrary function arguments
        if argtype == "KWARGS" and name in arguments:
            if vals[0] is not None:
                for key, item_vals in vals[0].items():  # Iterate over it like a dict
                    if not isinstance(item_vals, (list, tuple)):
                        item_vals = [item_vals]
                    newvals = []
                    for val in item_vals:
                        if val is None:
                            continue

                        if isinstance(val, _File):
                            newvals.append(val.path)
                            continue  # Don't copy artifacts

                        if not isinstance(val, str) or not os.path.exists(val):
                            newvals.append(val)
                            continue
                        val = os.path.abspath(val)

                        outname = os.path.join(inputs_dir, os.path.basename(val))
                        if os.path.exists(outname) and val != original_paths[outname]:
                            i = 0
                            while os.path.exists(outname):
                                parts = os.path.splitext(os.path.basename(val))
                                outname = os.path.join(
                                    inputs_dir, f"{parts[0]}_{i}{parts[1]}"
                                )
                                i += 1

                        original_paths[outname] = val

                        if "PM_SYMLINK" in os.environ:
                            os.symlink(val, outname)
                        else:
                            if os.path.isdir(val):
                                shutil.copytree(val, outname)
                            else:
                                shutil.copy(val, outname)
                        newvals.append(os.path.relpath(outname, write_dir))

                    if len(newvals) == 0:
                        arguments[name][key] = None
                    elif len(newvals) == 1:
                        arguments[name][key] = newvals[0]
                    else:
                        arguments[name][key] = newvals

    with open(os.path.join(inputs_dir, "inputs.json"), "w") as f:
        json.dump(
            {"function": function, "slpm_path": slpm_path, "arguments": arguments},
            f,
            indent=4,
        )

    with open(os.path.join(inputs_dir, "original_paths.json"), "w") as f:
        json.dump(original_paths, f, indent=4)


def _docs_from_manifest(manifest, appname, appdir):
    from copy import deepcopy

    manifest = deepcopy(manifest)

    if "description" not in manifest:
        raise RuntimeError(
            "Missing the 'description' field in your app manifest with a description of the app."
        )

    docs = [manifest["description"], "", "Parameters", "----------"]
    for i, param in enumerate(manifest["params"]):
        pp = f"{param['name']} : {param['type']}"
        if "choices" in param and param["choices"] is not None:
            choices = '", "'.join(param["choices"])
            pp += f', choices=("{choices}")'
        docs.append(pp)
        docs.append(f"    {param['description']}")

    missing = []

    if "outputs" in manifest:
        docs.append("")
        docs.append("Outputs")
        docs.append("-------")
        for key, val in manifest["outputs"].items():
            docs.append(key)
            docs.append(f"    {val}")
    else:
        missing.append("outputs")

    if "resources" in manifest and manifest["resources"] is not None:
        docs.append("")
        docs.append("Note")
        docs.append("----")
        docs.append("Minimum job requirements::")
        docs.append("")
        for key, val in manifest["resources"].items():
            docs.append(f"    {key}: {val}")

    if "examples" in manifest:
        docs.append("")
        docs.append("Examples")
        docs.append("--------")
        for exp in manifest["examples"]:
            docs.append(f">>> {exp}")
    else:
        missing.append("examples")

    if "tests" in manifest:
        for test_name in manifest["tests"]:
            desc = manifest["tests"][test_name]["description"]
            args = manifest["tests"][test_name]["arguments"]
            args_str = ""
            for key, vals in args.items():
                if not isinstance(vals, (list, tuple)):
                    vals = [vals]
                for i in range(len(vals)):
                    val = vals[i]
                    if isinstance(val, str):
                        if val.startswith("tests/"):
                            val = f"{appname}.files['{val}']"
                        elif val.startswith("datasets/"):
                            val = val.replace("datasets/", "")
                            val = f"{appname}.datasets.{val}"
                        elif val.startswith("artifacts/"):
                            val = val.replace("artifacts/", "")
                            val = f"{appname}.artifacts.{val}"
                        else:
                            val = f"'{val}'"
                    vals[i] = val
                if len(vals) > 1:
                    args_str += f"{key}=[{', '.join(map(str, vals))}], "
                else:
                    args_str += f"{key}={vals[0]}, "

            docs.append("")
            docs.append(desc)
            docs.append(f">>> {appname}(outdir='./out', {args_str[:-2]}).run()")

        if "examples" in missing:  # if there are tests don't complain about examples
            missing.remove("examples")

    if appdir is not None:
        tutorials = glob(os.path.join(appdir, "tutorials", "*"))
        if len(tutorials):
            docs.append("")
            docs.append("Notes")
            docs.append("-----")
            docs.append("Tutorials are available for this app:")
            docs.append("")
            for fname in tutorials:
                docs.append(f"    - {fname}")

    if len(missing):
        logger.warning(f"{appname} manifest is missing fields: {', '.join(missing)}")
    return docs


def _args_from_manifest(func_args):
    # Fix for old apps
    fix_old_types = {"string": "str", "file": "Path"}

    # Arguments which don't have a "value" field should be mandatory
    for arg in func_args:
        if "mandatory" not in arg:
            arg["mandatory"] = "value" not in arg

    args = []
    # Ensure mandaroty args come first if someone messes up the manifest
    mand_params = [x for x in func_args if x["mandatory"]]
    opt_params = [x for x in func_args if not x["mandatory"]]

    params = mand_params + opt_params
    for i, param in enumerate(params):
        atype = param["type"]
        if atype in fix_old_types:
            atype = fix_old_types[atype]
        if atype == "str_to_bool":
            atype = "bool"

        atype_final = atype
        if "nargs" in param and param["nargs"] is not None:
            atype_final = f"list[{atype}]"

        argstr = f"{param['name']} : {atype_final}"

        if not param["mandatory"]:
            default = param["value"]

            if atype in ("str", "Path") and param["value"] is not None:
                if isinstance(default, (list, tuple)):
                    for k in range(len(default)):
                        if default[k].startswith("app://files"):
                            # Handle app://files URIs
                            default[k] = (
                                f"files['{default.replace('app://files/', '')}']"
                            )
                elif default.startswith("app://files"):
                    # Handle app://files URIs
                    default = f"files['{default.replace('app://files/', '')}']"
                else:
                    default = f"\"{param['value']}\""

            # Fix for old apps
            if atype not in ("str", "Path") and param["value"] == "":
                default = None

            argstr += f" = {default}"

        if i != len(params) - 1:
            argstr += ","
        args.append(argstr)
    return args


def __ensure_module_path(module_path):
    """
    Ensures that all parts of the dotted module path exist in sys.modules
    and returns the final module object.
    """
    import sys
    import types

    parts = module_path.split(".")
    full_path = ""
    parent = None

    for part in parts:
        full_path = f"{full_path}.{part}" if full_path else part
        if full_path not in sys.modules:
            mod = types.ModuleType(full_path)
            sys.modules[full_path] = mod
            if parent:
                setattr(parent, part, mod)
        parent = sys.modules[full_path]

    return sys.modules[full_path]


def _manifest_to_func(appname, app_versions):
    from copy import deepcopy

    template = env.get_template("func.py.jinja")

    for version in app_versions:
        manifest = app_versions[version]["manifest"]
        new_mode = True
        if (
            "functions" not in manifest
        ):  # TODO: Deprecate eventually. Just for backwards compatibility
            new_mode = False
            manifest = deepcopy(manifest)
            try:
                manifest["functions"] = [
                    {
                        "function": (
                            manifest["container_config"]["appfunction"]
                            if "container_config" in manifest
                            else "main"
                        ),
                        "env": "base",
                        "resources": manifest.get("resources", None),
                        "examples": manifest.get("examples", []),
                        "params": manifest["params"],
                        "tests": manifest.get("tests", {}),
                        "outputs": manifest.get("outputs", {}),
                        "description": manifest["description"],
                    }
                ]
            except Exception:
                import traceback

                logger.error(
                    f"Failed to parse manifest for app {appname} version {version} due to error:\n{traceback.format_exc()}"
                )
                continue

        module_path = f"playmolecule.apps.{appname}.{version}"
        module = __ensure_module_path(module_path)

        setattr(module, "__manifest__", deepcopy(manifest))

        if "run.sh" in app_versions[version]:
            with open(app_versions[version].get("run.sh"), "r") as f:
                run_sh = f.read()
            setattr(module, "__runsh", run_sh)
        else:
            setattr(module, "__runsh", None)

        app_artifacts = None
        source_dir = app_versions[version].get("appdir")
        if source_dir is not None:
            source_dir = os.path.join(source_dir, "files")
        app_files = _get_app_files(source_dir, manifest)
        setattr(module, "files", app_files)

        artifact_dict = manifest.get("artifacts", manifest.get("datasets", {}))
        if len(artifact_dict):
            app_artifacts = _Artifacts(artifact_dict, app_files)
            setattr(module, "artifacts", app_artifacts)
            setattr(module, "datasets", app_artifacts)

        func_names = []
        for idx, func_mani in enumerate(manifest["functions"]):
            try:
                func_name = func_mani["function"].split(".")[-1]
                if func_name == "main":
                    func_name = appname

                code = template.render(
                    function=func_mani["function"],
                    function_idx=idx,
                    function_name=func_name,
                    function_args=_args_from_manifest(func_mani["params"]),
                    function_docs=_docs_from_manifest(
                        func_mani, appname, app_versions[version]["appdir"]
                    ),
                    function_resources=func_mani.get("resources", None),
                    module_path=module_path,
                    slpm_path=f"{appname}.{version}.{func_name}",
                )
                local_ns = {"files": app_files}
                exec(code, globals().copy(), local_ns)
                _func = local_ns[func_name]

                tests = _Tests(func_mani["tests"], _func, app_files, app_artifacts)
                # Add some metadata to the function object
                _func.tests = tests
                _func._name = func_name
                _func.__manifest__ = deepcopy(func_mani)
                func_names.append(func_name)
                setattr(module, func_name, _func)
                _function_dict[f"{module_path}.{func_name}"] = _func
            except Exception:
                import traceback

                logger.error(
                    f"Failed to parse manifest for app {appname} version {version} with error: {traceback.format_exc()}"
                )
    return func_names


def _link_latest_version(appname, latest, func_names):
    import sys

    # Link the latest version of the app to the top level module
    parent_module = sys.modules[f"playmolecule.apps.{appname}"]
    latest_module = sys.modules[f"playmolecule.apps.{appname}.{latest}"]
    for symbol in func_names + [
        "artifacts",
        "datasets",
        "files",
        "tests",
        "__manifest__",
    ]:
        if symbol not in parent_module.__dict__ and symbol in latest_module.__dict__:
            setattr(
                sys.modules[f"playmolecule.apps.{appname}"],
                symbol,
                getattr(
                    sys.modules.get(f"playmolecule.apps.{appname}.{latest}"), symbol
                ),
            )


def _check_folder_validity(root_dir, app_d):
    import re

    folder_path = os.path.relpath(app_d, root_dir)

    # Check if the folder consists only of lowercase letters, numbers, underscores or path separators
    if not re.match(r"^[a-z0-9_/\\]+$", folder_path):
        logger.warning(
            f'Path {app_d} has invalid characters in the part: "{folder_path}". Only lowercase letters, numbers and underscores are allowed. Please fix the path. Skipping...'
        )
        return False
    return True


def _set_root(root_dir):
    import json
    from natsort import natsorted
    import playmolecule

    if root_dir.startswith("http"):
        from playmolecule._executors._pmbackend import _set_root_pmbackend

        _set_root_pmbackend()
        return

    if root_dir.startswith("docker"):
        from playmolecule._docker import _set_docker_root_gcloud

        _set_docker_root_gcloud()
        return

    logger.info(f"PlayMolecule home: {root_dir}")

    _setup_folders(root_dir)

    for app_d in natsorted(glob(os.path.join(root_dir, "apps", "*", ""))):
        appname = os.path.basename(os.path.abspath(app_d))
        if not _check_folder_validity(root_dir, app_d):
            continue

        versions = glob(os.path.join(app_d, "*"))
        versions.sort(key=lambda s: natsorted(os.path.basename(s)))

        app_versions = {}
        for vv in versions:
            vname = os.path.basename(vv)
            if not _check_folder_validity(root_dir, vv):
                continue

            jf = glob(os.path.join(vv, "*.json"))
            if len(jf) > 1:
                print(f"ERROR: Multiple json files found in {vv}")
            if len(jf) == 1 and os.stat(jf[0]).st_size != 0:
                try:
                    with open(jf[0]) as f:
                        app_versions[vname] = {
                            "manifest": json.load(f),
                            "appdir": vv,
                            "run.sh": os.path.join(vv, "run.sh"),
                        }
                except Exception as e:
                    logger.error(
                        f"Failed at parsing manifest JSON file {jf[0]} with error: {e}"
                    )
        if len(app_versions):
            func_names = _manifest_to_func(appname, app_versions)
            _link_latest_version(appname, os.path.basename(versions[-1]), func_names)

    dsdir = os.path.join(root_dir, "datasets")
    dsjson = os.path.join(dsdir, "datasets.json")
    if os.path.exists(dsdir) and os.path.exists(dsjson):
        with open(dsjson) as f:
            manifest = json.load(f)
        files = _get_app_files(dsdir)
        playmolecule.datasets = _Artifacts(manifest["datasets"], files)


def _setup_folders(root_dir):
    os.makedirs(root_dir, exist_ok=True)
    os.makedirs(os.path.join(root_dir, "apps"), exist_ok=True)
    os.makedirs(os.path.join(root_dir, "datasets"), exist_ok=True)

    apptainer_runner = os.path.join(root_dir, "apptainer_run.sh")
    if not os.path.exists(apptainer_runner):
        try:
            import questionary

            license_type = questionary.select(
                "Do you use a floating license server (IP/Port) or a license file?:",
                choices=["Floating License Server", "License File"],
                default="Floating License Server",
                use_shortcuts=True,
            ).unsafe_ask()
            if license_type == "Floating License Server":
                license_ip = questionary.text(
                    message="Type the IP/URL of the license server:"
                ).unsafe_ask()
                license_port = questionary.text(
                    message="Type the port of the license server:", default="27000"
                ).unsafe_ask()
                license_file = f"{license_port}@{license_ip}"
            else:
                license_file = questionary.path(
                    message="Path to license file:",
                ).unsafe_ask()
                new_lic_file = os.path.join(root_dir, "license.dat")
                if os.path.abspath(new_lic_file) != os.path.abspath(license_file):
                    shutil.copy(license_file, new_lic_file)
                license_file = new_lic_file
        except KeyboardInterrupt:
            raise RuntimeError("PlayMolecule setup cancelled...")

        template = env.get_template("apptainer_run.sh")
        fstring = template.render(
            license_file_or_server=license_file,
            root_dir=root_dir,
        )
        with open(apptainer_runner, "w") as f:
            f.write(fstring)

        st = os.stat(apptainer_runner)
        os.chmod(
            apptainer_runner, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
        )

        with open(os.path.join(root_dir, "license_path.txt"), "w") as f:
            f.write(license_file)

    if len(glob(os.path.join(root_dir, "apps", "*", ""))) == 0:
        from playmolecule._update import update_apps

        update_apps()



from pathlib import Path
import os


class SleepException(Exception):
    pass


def report_alive(sentinel_dir):
    from datetime import datetime
    import time

    sentinel_dir = os.path.abspath(sentinel_dir)
    os.makedirs(sentinel_dir, exist_ok=True)

    while True:
        try:
            with open(os.path.join(sentinel_dir, ".pm.alive"), "w") as f:
                f.write(datetime.now().isoformat())
        except Exception:
            pass
        time.sleep(5)


def dump_manifest(func, outfile, files_path=None):
    import json
    import importlib

    pieces = func.split(".")
    module_name = ".".join(pieces[:-1])
    module = importlib.import_module(module_name)
    manifest = module.__manifest__

    if files_path is not None:
        manifest["files"] = get_file_list(files_path)

    with open(outfile, "w") as f:
        json.dump(manifest, f)


def get_file_list(files_path):
    from glob import glob
    import os

    files = {
        os.path.relpath(x, files_path): x
        for x in glob(os.path.join(files_path, "**", "*"), recursive=True)
        if os.path.isfile(x)
    }

    return files


def app_wrapper(func=None, cli_arg_str=None, input_json=None):
    from func2argparse import func_to_manifest, manifest_to_argparser
    from unittest import mock
    from shlex import split
    import importlib
    import threading
    import shutil
    import json

    json_dir = None
    if input_json is not None:
        json_dir = os.path.dirname(os.path.abspath(input_json))
        with open(input_json, "r") as f:
            params = json.load(f)
        if "function" in params:
            func = params["function"]
            params = params["arguments"]

    # Import the module/function of the app to call
    pieces = func.split(".")
    module_name = ".".join(pieces[:-1])
    module = importlib.import_module(module_name)
    func = getattr(module, pieces[-1])

    if input_json is None:
        manifest = func_to_manifest(func)
        parser = manifest_to_argparser(manifest)
        cli_arg = split(cli_arg_str)
        print("Parsing arguments", cli_arg)
        # The parser really likes to kill the app when failing to parse... This mock.patch prevents it
        with mock.patch("sys.exit") as m:
            appargs, unknownargs = parser.parse_known_args(cli_arg)
            if len(unknownargs):
                # Otherwise throw an error for unknown arguments
                parser.print_help()
                raise RuntimeError(
                    f"Unrecognized arguments \"{' '.join(unknownargs)}\""
                )
            if m.call_count > 0:
                raise RuntimeError("Failed parsing app arguments")
        params = vars(appargs)

    # Convert empty strings to Nones and Paths to strings
    for arg, value in params.items():
        if value == "":
            params[arg] = None
        if isinstance(value, Path):
            params[arg] = str(value)
        if isinstance(value, list) or isinstance(value, tuple):
            params[arg] = [str(x) if isinstance(x, Path) else x for x in value]

    if "outdir" in params:
        sentinel_dir = params["outdir"]
    elif "execdir" in params:
        sentinel_dir = params["execdir"]
    else:
        raise RuntimeError("No outdir or execdir specified in the arguments")

    if json_dir is not None:
        # If the input_json is provided, use the directory of the input_json as the sentinel_dir
        sentinel_dir = json_dir
    os.makedirs(sentinel_dir, exist_ok=True)

    scratchdir = params.get("scratchdir", None)
    if scratchdir is not None:
        if input_json is not None:
            # Create a scratchdir with a unique identifier based on the input_json path
            identifier = os.path.basename(json_dir)
            scratchdir = os.path.join(params["scratchdir"], identifier)

        params["scratchdir"] = scratchdir
        os.makedirs(scratchdir, exist_ok=True)

    thread = threading.Thread(target=report_alive, args=(sentinel_dir,), daemon=True)
    thread.start()

    try:
        func(**params)
    except SleepException:
        with open(os.path.join(sentinel_dir, ".pm.sleep"), "w") as f:
            f.write("")
    except Exception as e:
        with open(os.path.join(sentinel_dir, ".pm.err"), "w") as f:
            f.write("")
        raise e
    else:
        with open(os.path.join(sentinel_dir, ".pm.done"), "w") as f:
            f.write("")
        # Check that scratchdir is not the whole home or tmp directory or /
        if scratchdir is not None:
            scratchdir = Path(scratchdir).resolve()
            if scratchdir not in (Path.home().resolve(), Path("/tmp"), Path("/")):
                shutil.rmtree(scratchdir)
    return


def ensurelist(tocheck, tomod=None):
    """Convert np.ndarray and scalars to lists.

    Lists and tuples are left as is. If a second argument is given,
    the type check is performed on the first argument, and the second argument is converted.
    """
    if tomod is None:
        tomod = tocheck
    if type(tocheck).__name__ == "ndarray":
        return list(tomod)
    if isinstance(tocheck, range):
        return list(tocheck)
    if not isinstance(tocheck, list) and not isinstance(tocheck, tuple):
        return [
            tomod,
        ]
    return tomod


def wait_jobs(dirs, name, wait, logger, error=True, sleep_time=30):
    """
    Wait for jobs to complete or return directly if any job has not completed
    """
    from playmolecule.apps import ExecutableDirectory, JobStatus
    import time

    dirs = ensurelist(dirs)
    if len(dirs) == 0:
        raise RuntimeError(
            f"No {name} directories to wait for. Check if the application ran correctly."
        )

    while True:
        non_completed = [
            ExecutableDirectory(dd).status not in (JobStatus.COMPLETED, JobStatus.ERROR)
            for dd in dirs
        ]

        if error and any(
            ExecutableDirectory(dd).status == JobStatus.ERROR for dd in dirs
        ):
            raise RuntimeError(f"Error in {name} job")

        if any(non_completed) and wait:
            logger.info(
                f"Waiting for {sum(non_completed)} {name} job to complete. Sleeping for {sleep_time}s"
            )
            time.sleep(sleep_time)
        else:
            break
    return any(non_completed)



from glob import glob
import json
import os
import tempfile
import pytest
import sys

curr_dir = os.path.dirname(os.path.abspath(__file__))


def prepare(datadir):
    outdir = str(datadir.join("out"))
    scratchdir = str(datadir.join("scratch"))
    os.makedirs(outdir, exist_ok=True)
    os.makedirs(scratchdir, exist_ok=True)
    return outdir, scratchdir


@pytest.fixture
def test_app_root():
    """Set up the test app root"""
    old_root = os.environ.get("PM_APP_ROOT")
    os.environ["PM_APP_ROOT"] = os.path.join(curr_dir, "test_playmolecule")
    yield os.path.join(curr_dir, "test_playmolecule")
    if old_root:
        os.environ["PM_APP_ROOT"] = old_root
    else:
        os.environ.pop("PM_APP_ROOT", None)


def _test_old_playmolecule_manifests():
    os.environ["PM_APP_ROOT"] = os.path.join(curr_dir, "test_playmolecule")

    from playmolecule import apps, datasets, protocols
    import tempfile

    assert hasattr(apps, "proteinprepare")
    assert hasattr(apps.proteinprepare, "files")
    assert hasattr(apps.proteinprepare, "proteinprepare")
    assert hasattr(apps.proteinprepare.v1, "proteinprepare")
    assert hasattr(apps.proteinprepare.v1, "files")
    assert hasattr(apps.proteinprepare.v1.proteinprepare, "tests")
    assert hasattr(apps.proteinprepare.v1.proteinprepare.tests, "simple")
    assert sorted(list(apps.proteinprepare.v1.files.keys())) == sorted(
        [
            "datasets",
            "datasets/3ptb.pdb",
            "tests",
            "tests/web_content.pickle",
            "tests/reprepare.pickle",
            "tests/3ptb.pdb",
            "tests/587HG92V.pdb",
            "tutorials",
            "tutorials/learn_this_app.ipynb",
        ]
    )
    assert hasattr(apps.proteinprepare.v1.datasets, "file_3ptb")
    assert hasattr(datasets, "file_3ptb")
    assert hasattr(protocols, "crypticscout")
    assert hasattr(protocols.crypticscout, "v1")
    assert hasattr(protocols.crypticscout.v1, "crypticscout")
    assert hasattr(protocols.crypticscout.v1.crypticscout, "crypticscout")
    assert callable(protocols.crypticscout.v1.crypticscout.crypticscout)

    expected_files = [
        "run_*.sh",
        "run_*/",
        "run_*/expected_outputs.json",
        "run_*/.pm.done",
        os.path.join("run_*", "inputs.json"),
        os.path.join("run_*", "original_paths.json"),
    ]
    with tempfile.TemporaryDirectory() as tmpdir:
        apps.proteinprepare.proteinprepare(tmpdir, pdbfile=datasets.file_3ptb).run()
        for ef in expected_files:
            assert len(glob(os.path.join(tmpdir, ef))), "could not find " + ef

    expected_files += [os.path.join("run_*", "3ptb.pdb")]
    with tempfile.TemporaryDirectory() as tmpdir:
        apps.proteinprepare.proteinprepare(
            tmpdir,
            pdbfile=os.path.join(curr_dir, "test_playmolecule", "datasets", "3ptb.pdb"),
        ).run()
        for ef in expected_files:
            assert len(glob(os.path.join(tmpdir, ef))), "could not find " + ef


def _test_new_playmolecule_manifests():
    os.environ["PM_APP_ROOT"] = os.path.join(curr_dir, "test_playmolecule")

    from playmolecule import apps, datasets
    import tempfile

    assert hasattr(apps, "proteinpreparenew")
    assert hasattr(apps.proteinpreparenew, "proteinpreparenew")
    assert hasattr(apps.proteinpreparenew, "files")
    assert hasattr(apps.proteinpreparenew.v1, "proteinpreparenew")
    assert hasattr(apps.proteinpreparenew.v1, "files")
    assert hasattr(apps.proteinpreparenew.v1.proteinpreparenew, "tests")
    assert hasattr(apps.proteinpreparenew.v1.proteinpreparenew.tests, "simple")
    assert hasattr(apps.proteinpreparenew.v2, "proteinpreparenew")
    assert hasattr(apps.proteinpreparenew.v2, "files")
    assert hasattr(apps.proteinpreparenew.v2.proteinpreparenew, "tests")
    assert hasattr(apps.proteinpreparenew.v2.proteinpreparenew.tests, "simple")
    assert sorted(list(apps.proteinpreparenew.v1.files.keys())) == sorted(
        [
            "datasets",
            "datasets/3ptb.pdb",
            "tests",
            "tests/web_content.pickle",
            "tests/reprepare.pickle",
            "tests/3ptb.pdb",
            "tests/587HG92V.pdb",
            "tutorials",
            "tutorials/learn_this_app.ipynb",
        ]
    )
    assert hasattr(apps.proteinpreparenew.v1.datasets, "file_3ptb")
    assert hasattr(datasets, "file_3ptb")

    expected_files = [
        "run_*.sh",
        "run_*/",
        "run_*/.pm.done",
        "run_*/expected_outputs.json",
        os.path.join("run_*", "inputs.json"),
        os.path.join("run_*", "original_paths.json"),
    ]
    with tempfile.TemporaryDirectory() as tmpdir:
        apps.proteinpreparenew.proteinpreparenew(
            tmpdir, pdbfile=datasets.file_3ptb
        ).run()
        for ef in expected_files:
            assert len(glob(os.path.join(tmpdir, ef))), "could not find " + ef

        with open(glob(os.path.join(tmpdir, "run_*", "inputs.json"))[0], "r") as f:
            inputs = json.load(f)
            assert "function" in inputs
            assert (
                inputs["function"] == "proteinprepare.apps.proteinpreparenew.app.main"
            )

    with tempfile.TemporaryDirectory() as tmpdir:
        apps.proteinpreparenew.v1.bar(tmpdir, pdbid="3ptb").run()
        for ef in expected_files:
            assert len(glob(os.path.join(tmpdir, ef))), "could not find " + ef

        with open(glob(os.path.join(tmpdir, "run_*", "inputs.json"))[0], "r") as f:
            inputs = json.load(f)
            assert "function" in inputs
            assert inputs["function"] == "proteinprepare.apps.proteinpreparenew.app.bar"

    expected_files += [os.path.join("run_*", "3ptb.pdb")]
    with tempfile.TemporaryDirectory() as tmpdir:
        apps.proteinpreparenew.proteinpreparenew(
            tmpdir,
            pdbfile=os.path.join(curr_dir, "test_playmolecule", "datasets", "3ptb.pdb"),
        ).run()
        for ef in expected_files:
            assert len(glob(os.path.join(tmpdir, ef))), "could not find " + ef

    with tempfile.TemporaryDirectory() as tmpdir:
        apps.proteinpreparenew.proteinpreparenew(
            tmpdir, pdbfile="app://files/datasets/3ptb.pdb"
        ).run()
        assert (
            len(glob(os.path.join(tmpdir, "run_*", "3ptb.pdb"))) == 0
        ), "3ptb file should not be copied since it's a dataset"

        with open(glob(os.path.join(tmpdir, "run_*", "inputs.json"))[0], "r") as f:
            inputs = json.load(f)
            assert "function" in inputs
            assert (
                inputs["function"] == "proteinprepare.apps.proteinpreparenew.app.main"
            )
            assert (
                inputs["arguments"]["pdbfile"]
                == apps.proteinpreparenew.files["datasets/3ptb.pdb"].path
            )


# ============================================================================
# Test _appfiles module
# ============================================================================


def _test_file_class():
    """Test the _File class"""
    from playmolecule._appfiles import _File

    # Test basic file creation
    file = _File("test", "/path/to/file", "Test description")
    assert file.name == "test"
    assert file.path == "/path/to/file"
    assert file.description == "Test description"

    # Test string representation
    assert "[test]" in str(file)
    assert "/path/to/file" in str(file)
    assert "Test description" in str(file)

    # Test file without description
    file_no_desc = _File("test2", "/path/to/file2")
    assert file_no_desc.description is None
    assert "test2" in str(file_no_desc)


def _test_artifacts_class(test_app_root):
    """Test the _Artifacts class"""
    from playmolecule._appfiles import _Artifacts, _File

    files_dict = {
        "datasets/test.pdb": _File("test.pdb", "/path/to/test.pdb"),
        "artifacts/data.txt": _File("data.txt", "/path/to/data.txt"),
    }

    artifacts_list = [
        {"name": "file_test", "path": "test.pdb", "description": "Test file"},
        {"name": "file_data", "path": "data.txt", "description": "Data file"},
    ]

    artifacts = _Artifacts(artifacts_list, files_dict)

    # Test that attributes were created correctly
    assert hasattr(artifacts, "file_test")
    assert hasattr(artifacts, "file_data")

    # Test string representation
    str_repr = str(artifacts)
    assert "file_test" in str_repr
    assert "file_data" in str_repr


def _test_artifacts_validation(test_app_root):
    """Test _Artifacts validation"""
    from playmolecule._appfiles import _Artifacts, _File

    files_dict = {"datasets/test.pdb": _File("test.pdb", "/path/to/test.pdb")}

    # Test that names with dots are rejected
    artifacts_list = [
        {"name": "file.with.dots", "path": "test.pdb", "description": "Test"}
    ]
    artifacts = _Artifacts(artifacts_list, files_dict)
    assert not hasattr(artifacts, "file.with.dots")

    # Test that names starting with numbers are rejected
    artifacts_list = [{"name": "123file", "path": "test.pdb", "description": "Test"}]
    artifacts = _Artifacts(artifacts_list, files_dict)
    assert not hasattr(artifacts, "123file")


def _test_get_app_files(test_app_root):
    """Test _get_app_files function"""
    from playmolecule._appfiles import _get_app_files

    source_dir = os.path.join(test_app_root, "apps", "proteinprepare", "v1", "files")
    files = _get_app_files(source_dir)

    # Check that files were found
    assert isinstance(files, dict)
    assert len(files) > 0

    # Check that keys are relative paths
    for key in files.keys():
        assert not os.path.isabs(key)


# ============================================================================
# Test _protocols module
# ============================================================================


def _test_find_protocols(test_app_root):
    """Test __find_protocols function"""
    from playmolecule._protocols import __find_protocols

    protocols = __find_protocols(test_app_root)
    assert protocols is not None

    # Check that the protocols module was created
    assert hasattr(protocols, "crypticscout")
    assert hasattr(protocols.crypticscout, "v1")


def _test_find_protocols_nonexistent():
    """Test __find_protocols with non-existent directory"""
    from playmolecule._protocols import __find_protocols

    protocols = __find_protocols("/nonexistent/path")
    assert protocols is None


# ============================================================================
# Test _tests module
# ============================================================================


def _test_tests_class(test_app_root):
    """Test _Tests class"""
    os.environ["PM_APP_ROOT"] = test_app_root
    from playmolecule import apps
    from playmolecule._tests import _Tests

    # Get a function with tests
    func = apps.proteinprepare.v1.proteinprepare
    assert hasattr(func, "tests")
    assert hasattr(func.tests, "simple")

    # Test string representation
    tests_str = str(func.tests)
    assert "simple" in tests_str


def _test_test_class_attributes(test_app_root):
    """Test _Test class attributes"""
    os.environ["PM_APP_ROOT"] = test_app_root
    from playmolecule import apps

    func = apps.proteinprepare.v1.proteinprepare
    test = func.tests.simple

    assert hasattr(test, "func")
    assert hasattr(test, "name")
    assert hasattr(test, "config")
    assert callable(test.run)

    # Test string representation
    test_str = str(test)
    assert "Arguments" in test_str
    assert "Expected outputs" in test_str


def _test_convert_test_files(test_app_root):
    """Test _convert_test_files function"""
    from playmolecule._tests import _convert_test_files
    from playmolecule._appfiles import _File

    files = {
        "tests/test.pdb": _File("test.pdb", "/path/to/test.pdb"),
    }

    # Test with test file
    result = _convert_test_files("tests/test.pdb", files)
    assert isinstance(result, _File)
    assert result.path == "/path/to/test.pdb"

    # Test with regular string
    result = _convert_test_files("regular_string", files)
    assert result == "regular_string"


# ============================================================================
# Test apps module - JobStatus
# ============================================================================


def _test_job_status_enum():
    """Test JobStatus enum"""
    from playmolecule import JobStatus

    # Test enum values
    assert JobStatus.WAITING_INFO == 0
    assert JobStatus.RUNNING == 1
    assert JobStatus.COMPLETED == 2
    assert JobStatus.ERROR == 3

    # Test describe method
    assert JobStatus.WAITING_INFO.describe() == "Waiting info"
    assert JobStatus.RUNNING.describe() == "Running"
    assert JobStatus.COMPLETED.describe() == "Completed"
    assert JobStatus.ERROR.describe() == "Error"

    # Test string representation
    assert str(JobStatus.COMPLETED) == "Completed"


# ============================================================================
# Test apps module - ExecutableDirectory
# ============================================================================


def _test_executable_directory_init(test_app_root):
    """Test ExecutableDirectory initialization"""
    from playmolecule import apps
    from playmolecule.apps import ExecutableDirectory

    os.environ["PM_APP_ROOT"] = test_app_root
    from playmolecule import datasets

    with tempfile.TemporaryDirectory() as tmpdir:
        # Create an executable directory
        ed = apps.proteinprepare.proteinprepare(tmpdir, pdbfile=datasets.file_3ptb)

        assert ed.dirname == tmpdir
        assert ed.runsh is not None
        assert ed.runsh.endswith(".sh")
        assert ed.inputs_dir is not None


def _test_executable_directory_status_completed(test_app_root):
    """Test ExecutableDirectory status with completed job"""
    from playmolecule import apps, JobStatus

    os.environ["PM_APP_ROOT"] = test_app_root
    from playmolecule import datasets

    with tempfile.TemporaryDirectory() as tmpdir:
        ed = apps.proteinprepare.proteinprepare(tmpdir, pdbfile=datasets.file_3ptb)
        ed.run()

        # Check status after completion
        assert ed.status == JobStatus.COMPLETED


def _test_executable_directory_status_error():
    """Test ExecutableDirectory status detection for errors"""
    from playmolecule.apps import ExecutableDirectory, JobStatus

    with tempfile.TemporaryDirectory() as tmpdir:
        # Create a fake run script
        runsh = os.path.join(tmpdir, "run_test.sh")
        with open(runsh, "w") as f:
            f.write("#!/bin/bash\n")

        inputs_dir = os.path.join(tmpdir, "run_test")
        os.makedirs(inputs_dir, exist_ok=True)

        # Create error marker
        with open(os.path.join(inputs_dir, ".pm.err"), "w") as f:
            f.write("")

        ed = ExecutableDirectory(tmpdir, inputs_dir=inputs_dir)
        assert ed.status == JobStatus.ERROR

        ed = ExecutableDirectory(tmpdir)
        assert ed.status == JobStatus.ERROR


# ============================================================================
# Test apps module - function creation and validation
# ============================================================================


def _test_validators():
    """Test type validators"""
    from playmolecule.apps import _validators
    from pathlib import Path
    from playmolecule._appfiles import _File

    # Test str validator
    assert isinstance("test", _validators["str"])

    # Test Path validator
    assert isinstance("/path/to/file", _validators["Path"])
    assert isinstance(Path("/path"), _validators["Path"])
    assert isinstance(_File("test", "/path"), _validators["Path"])

    # Test bool validator
    assert isinstance(True, _validators["bool"])

    # Test int validator
    assert isinstance(1, _validators["int"])
    assert isinstance(1.0, _validators["int"])

    # Test float validator
    assert isinstance(1.0, _validators["float"])

    # Test dict validator
    assert isinstance({}, _validators["dict"])


def _test_write_inputs(test_app_root):
    """Test _write_inputs function"""
    from playmolecule.apps import _write_inputs
    from playmolecule._appfiles import _get_app_files

    os.environ["PM_APP_ROOT"] = test_app_root

    with tempfile.TemporaryDirectory() as tmpdir:
        inputdir = os.path.join(tmpdir, "run_test")

        # Simple manifest for testing
        manifest = {
            "params": [
                {"name": "outdir", "type": "str"},
                {"name": "value", "type": "int"},
            ],
            "outputs": {"output.txt": "Test output"},
        }

        arguments = {"outdir": ".", "value": 42}
        app_files = {}

        _write_inputs(
            tmpdir,
            inputdir,
            arguments,
            manifest,
            app_files,
            "test.func",
            "test.v1.func",
        )

        # Check that files were created
        assert os.path.exists(os.path.join(inputdir, "inputs.json"))
        assert os.path.exists(os.path.join(inputdir, "expected_outputs.json"))
        assert os.path.exists(os.path.join(inputdir, "original_paths.json"))

        # Check inputs.json content
        with open(os.path.join(inputdir, "inputs.json"), "r") as f:
            inputs = json.load(f)
            assert inputs["function"] == "test.func"
            assert inputs["arguments"]["value"] == 42


def _test_args_from_manifest():
    """Test _args_from_manifest function"""
    from playmolecule.apps import _args_from_manifest

    # Test mandatory parameters
    func_args = [
        {"name": "param1", "type": "str", "mandatory": True},
        {"name": "param2", "type": "int", "value": 10, "mandatory": False},
    ]

    args = _args_from_manifest(func_args)
    assert len(args) == 2
    assert "param1 : str," in args[0]
    assert "param2 : int = 10" in args[1]


def _test_docs_from_manifest():
    """Test _docs_from_manifest function"""
    from playmolecule.apps import _docs_from_manifest

    manifest = {
        "description": "Test app description",
        "params": [
            {"name": "input", "type": "Path", "description": "Input file"},
            {"name": "output", "type": "str", "description": "Output file"},
        ],
        "outputs": {"result.txt": "Result file"},
        "resources": {"ncpu": 1, "ngpu": 0},
    }

    docs = _docs_from_manifest(manifest, "testapp", None)

    assert len(docs) > 0
    assert "Test app description" in docs
    assert "Parameters" in docs
    assert "Outputs" in docs
    assert "Note" in docs


def _test_check_folder_validity():
    """Test _check_folder_validity function"""
    from playmolecule.apps import _check_folder_validity

    root = "/home/test"

    # Valid folder names
    assert _check_folder_validity(root, "/home/test/app_name")
    assert _check_folder_validity(root, "/home/test/app123")
    assert _check_folder_validity(root, "/home/test/my_app")

    # Invalid folder names
    assert not _check_folder_validity(root, "/home/test/App-Name")
    assert not _check_folder_validity(root, "/home/test/app.name")
    assert not _check_folder_validity(root, "/home/test/app name")


# ============================================================================
# Test _devutils module
# ============================================================================


def _test_ensurelist():
    """Test ensurelist function"""
    from playmolecule._devutils import ensurelist

    # Test with list
    assert ensurelist([1, 2, 3]) == [1, 2, 3]

    # Test with tuple
    assert ensurelist((1, 2, 3)) == (1, 2, 3)

    # Test with scalar
    assert ensurelist(1) == [1]

    # Test with string
    assert ensurelist("test") == ["test"]

    # Test with range
    assert ensurelist(range(3)) == [0, 1, 2]


def _test_ensurelist_with_tomod():
    """Test ensurelist with separate check and modification values"""
    from playmolecule._devutils import ensurelist

    # Test with tocheck as list, tomod as different value
    result = ensurelist([1, 2], "value")
    assert result == "value"

    # Test with tocheck as scalar, tomod as different value
    result = ensurelist(1, "value")
    assert result == ["value"]


def _test_report_alive():
    """Test report_alive function runs in thread"""
    from playmolecule._devutils import report_alive
    import threading
    import time

    with tempfile.TemporaryDirectory() as tmpdir:
        inputs_dir = os.path.join(tmpdir, "test_sentinel")

        # Start the thread
        thread = threading.Thread(target=report_alive, args=(inputs_dir,), daemon=True)
        thread.start()

        # Give it a moment to write the file
        time.sleep(0.5)

        # Check that the alive file was created
        assert os.path.exists(os.path.join(inputs_dir, ".pm.alive"))

        # Read the file and verify it contains a timestamp
        with open(os.path.join(inputs_dir, ".pm.alive"), "r") as f:
            content = f.read()
            assert len(content) > 0


def _test_dump_manifest(test_app_root):
    """Test dump_manifest function"""
    from playmolecule._devutils import dump_manifest

    os.environ["PM_APP_ROOT"] = test_app_root
    # Need to import apps after setting the environment variable
    from playmolecule import apps

    with tempfile.TemporaryDirectory() as tmpdir:
        outfile = os.path.join(tmpdir, "manifest.json")

        # This should work with the test apps
        dump_manifest("playmolecule.apps.proteinprepare.v1", outfile)

        assert os.path.exists(outfile)

        with open(outfile, "r") as f:
            manifest = json.load(f)
            assert "functions" in manifest


# ============================================================================
# Test main module integration
# ============================================================================


def _test_describe_apps(test_app_root, capsys):
    """Test describe_apps function"""
    os.environ["PM_APP_ROOT"] = test_app_root

    # Need to reload the module to pick up the new PM_APP_ROOT
    if "playmolecule" in sys.modules:
        del sys.modules["playmolecule"]
    if "playmolecule.apps" in sys.modules:
        del sys.modules["playmolecule.apps"]

    from playmolecule import describe_apps

    describe_apps()
    captured = capsys.readouterr()

    # Check that output was produced
    assert len(captured.out) > 0


def _test_version():
    """Test version import"""
    from playmolecule import __version__

    # Version should be set (either from package or not set)
    assert __version__ is not None or __version__ == "__version"


def _test_pm_skip_setup():
    """Test PM_SKIP_SETUP environment variable"""
    old_skip = os.environ.get("PM_SKIP_SETUP")
    old_root = os.environ.get("PM_APP_ROOT")

    try:
        os.environ["PM_SKIP_SETUP"] = "1"
        if "PM_APP_ROOT" in os.environ:
            del os.environ["PM_APP_ROOT"]

        # Need to reload the module
        if "playmolecule" in sys.modules:
            del sys.modules["playmolecule"]

        # This should not raise an error even without PM_APP_ROOT
        import playmolecule

        assert playmolecule.PM_SKIP_SETUP

    finally:
        if old_skip:
            os.environ["PM_SKIP_SETUP"] = old_skip
        else:
            os.environ.pop("PM_SKIP_SETUP", None)

        if old_root:
            os.environ["PM_APP_ROOT"] = old_root


# ============================================================================
# Test KWARGS class
# ============================================================================


def _test_kwargs_class():
    """Test KWARGS class"""
    from playmolecule.apps import KWARGS

    kwargs = KWARGS({"a": 1, "b": 2})
    assert isinstance(kwargs, dict)
    assert kwargs["a"] == 1
    assert kwargs["b"] == 2


# ============================================================================
# Test error handling
# ============================================================================


def _test_write_inputs_type_validation():
    """Test type validation in _write_inputs"""
    from playmolecule.apps import _write_inputs

    with tempfile.TemporaryDirectory() as tmpdir:
        inputdir = os.path.join(tmpdir, "run_test")

        manifest = {
            "params": [
                {"name": "value", "type": "int"},
            ],
        }

        # Pass wrong type
        arguments = {"value": "not_an_int"}
        app_files = {}

        with pytest.raises(RuntimeError, match="Was expecting value of type"):
            _write_inputs(
                tmpdir,
                inputdir,
                arguments,
                manifest,
                app_files,
                "test.func",
                "test.v1.func",
            )


def _test_write_inputs_nargs_validation():
    """Test nargs validation in _write_inputs"""
    from playmolecule.apps import _write_inputs

    with tempfile.TemporaryDirectory() as tmpdir:
        inputdir = os.path.join(tmpdir, "run_test")

        manifest = {
            "params": [
                {"name": "value", "type": "int", "nargs": None},
            ],
        }

        # Pass list when single value expected
        arguments = {"value": [1, 2, 3]}
        app_files = {}

        with pytest.raises(RuntimeError, match="Was expecting a single value"):
            _write_inputs(
                tmpdir,
                inputdir,
                arguments,
                manifest,
                app_files,
                "test.func",
                "test.v1.func",
            )


def _test_sleep_exception():
    """Test SleepException"""
    from playmolecule._devutils import SleepException

    with pytest.raises(SleepException):
        raise SleepException("Test exception")


# ============================================================================
# Test path handling
# ============================================================================


def _test_path_copying_in_write_inputs(test_app_root):
    """Test that Path-type arguments are copied correctly"""
    from playmolecule.apps import _write_inputs

    with tempfile.TemporaryDirectory() as tmpdir:
        # Create a test file
        test_file = os.path.join(tmpdir, "input.txt")
        with open(test_file, "w") as f:
            f.write("test content")

        inputdir = os.path.join(tmpdir, "run_test")

        manifest = {
            "params": [
                {"name": "inputfile", "type": "Path"},
            ],
        }

        arguments = {"inputfile": test_file}
        app_files = {}

        _write_inputs(
            tmpdir,
            inputdir,
            arguments,
            manifest,
            app_files,
            "test.func",
            "test.v1.func",
        )

        # Check that the file was copied
        copied_file = os.path.join(inputdir, "input.txt")
        assert os.path.exists(copied_file)

        # Check original_paths.json
        with open(os.path.join(inputdir, "original_paths.json"), "r") as f:
            orig_paths = json.load(f)
            assert test_file in orig_paths.values()


def _test_app_files_uri_handling():
    """Test handling of app://files URIs"""
    from playmolecule.apps import _write_inputs
    from playmolecule._appfiles import _File

    with tempfile.TemporaryDirectory() as tmpdir:
        inputdir = os.path.join(tmpdir, "run_test")

        manifest = {
            "params": [
                {"name": "datafile", "type": "Path"},
            ],
        }

        app_files = {
            "datasets/test.pdb": _File("test.pdb", "/actual/path/test.pdb"),
        }

        arguments = {"datafile": "app://files/datasets/test.pdb"}

        _write_inputs(
            tmpdir,
            inputdir,
            arguments,
            manifest,
            app_files,
            "test.func",
            "test.v1.func",
        )

        # Check that the URI was resolved
        with open(os.path.join(inputdir, "inputs.json"), "r") as f:
            inputs = json.load(f)
            # The app://files URI should be resolved to the actual path
            assert inputs["arguments"]["datafile"] == "/actual/path/test.pdb"
